# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test quaternion calculations"""

import numpy as np
import pytest
from numpy import pi
from numpy.testing import assert_array_almost_equal, assert_array_equal

from .. import eulerangles as nea
from .. import quaternions as nq


def norm(vec):
    # Return unit vector with same orientation as input vector
    return vec / np.sqrt(vec @ vec)


def gen_vec(dtype):
    # Generate random 3-vector in [-1, 1]^3
    rand = np.random.default_rng()
    return rand.uniform(low=-1.0, high=1.0, size=(3,)).astype(dtype)


# Example rotations
eg_rots = [
    nea.euler2mat(z, y, x)
    for z in np.arange(-pi, pi, pi / 2)
    for y in np.arange(-pi, pi, pi / 2)
    for x in np.arange(-pi, pi, pi / 2)
]

# Example quaternions (from rotations)
eg_quats = [nq.mat2quat(M) for M in eg_rots]
# M, quaternion pairs
eg_pairs = list(zip(eg_rots, eg_quats))

# Set of arbitrary unit quaternions
unit_quats = set(
    tuple(norm(np.r_[w, x, y, z]))
    for w in range(-2, 3)
    for x in range(-2, 3)
    for y in range(-2, 3)
    for z in range(-2, 3)
    if (w, x, y, z) != (0, 0, 0, 0)
)


def test_fillpos():
    # Takes np array
    xyz = np.zeros((3,))
    w, x, y, z = nq.fillpositive(xyz)
    assert w == 1
    # Or lists
    xyz = [0] * 3
    w, x, y, z = nq.fillpositive(xyz)
    assert w == 1
    # Errors with wrong number of values
    with pytest.raises(ValueError):
        nq.fillpositive([0, 0])
    with pytest.raises(ValueError):
        nq.fillpositive([0] * 4)
    # Errors with negative w2
    with pytest.raises(ValueError):
        nq.fillpositive([1.0] * 3)
    # Test corner case where w is near zero
    wxyz = nq.fillpositive([1, 0, 0])
    assert wxyz[0] == 0.0


@pytest.mark.parametrize('dtype', ('f4', 'f8'))
def test_fillpositive_plus_minus_epsilon(dtype):
    # Deterministic test for fillpositive threshold
    # We are trying to fill (x, y, z) with a w such that |(w, x, y, z)| == 1
    # If |(x, y, z)| is slightly off one, w should still be 0
    nptype = np.dtype(dtype).type

    # Obviously, |(x, y, z)| == 1
    baseline = np.array([0, 0, 1], dtype=dtype)

    # Obviously, |(x, y, z)| ~ 1
    plus = baseline * nptype(1 + np.finfo(dtype).eps)
    minus = baseline * nptype(1 - np.finfo(dtype).eps)

    assert nq.fillpositive(plus)[0] == 0.0
    assert nq.fillpositive(minus)[0] == 0.0

    # |(x, y, z)| > 1, no real solutions
    plus = baseline * nptype(1 + 2 * np.finfo(dtype).eps)
    with pytest.raises(ValueError):
        nq.fillpositive(plus)

    # |(x, y, z)| < 1, two real solutions, we choose positive
    minus = baseline * nptype(1 - 2 * np.finfo(dtype).eps)
    assert nq.fillpositive(minus)[0] > 0.0


@pytest.mark.parametrize('dtype', ('f4', 'f8'))
def test_fillpositive_simulated_error(dtype):
    # Nondeterministic test for fillpositive threshold
    # Create random vectors, normalize to unit length, and count on floating point
    # error to result in magnitudes larger/smaller than one
    # This is to simulate cases where a unit quaternion with w == 0 would be encoded
    # as xyz with small error, and we want to recover the w of 0

    # Permit 1 epsilon per value (default, but make explicit here)
    w2_thresh = 3 * np.finfo(dtype).eps

    for _ in range(50):
        xyz = norm(gen_vec(dtype))

        assert nq.fillpositive(xyz, w2_thresh)[0] == 0.0


def test_conjugate():
    # Takes sequence
    cq = nq.conjugate((1, 0, 0, 0))
    # Returns float type
    assert cq.dtype.kind == 'f'


def test_quat2mat():
    # also tested in roundtrip case below
    M = nq.quat2mat([1, 0, 0, 0])
    assert_array_almost_equal, M, np.eye(3)
    M = nq.quat2mat([3, 0, 0, 0])
    assert_array_almost_equal, M, np.eye(3)
    M = nq.quat2mat([0, 1, 0, 0])
    assert_array_almost_equal, M, np.diag([1, -1, -1])
    M = nq.quat2mat([0, 2, 0, 0])
    assert_array_almost_equal, M, np.diag([1, -1, -1])
    M = nq.quat2mat([0, 0, 0, 0])
    assert_array_almost_equal, M, np.eye(3)


def test_inverse_0():
    # Takes sequence
    iq = nq.inverse((1, 0, 0, 0))
    # Returns float type
    assert iq.dtype.kind == 'f'


@pytest.mark.parametrize(('M', 'q'), eg_pairs)
def test_inverse_1(M, q):
    iq = nq.inverse(q)
    iqM = nq.quat2mat(iq)
    iM = np.linalg.inv(M)
    assert np.allclose(iM, iqM)


def test_eye():
    qi = nq.eye()
    assert qi.dtype.kind == 'f'
    assert np.all([1, 0, 0, 0] == qi)
    assert np.allclose(nq.quat2mat(qi), np.eye(3))


def test_norm():
    qi = nq.eye()
    assert nq.norm(qi) == 1
    assert nq.isunit(qi)
    qi[1] = 0.2
    assert not nq.isunit(qi)


@pytest.mark.parametrize(('M1', 'q1'), eg_pairs[0::4])
@pytest.mark.parametrize(('M2', 'q2'), eg_pairs[1::4])
def test_mult(M1, q1, M2, q2):
    # Test that quaternion * same as matrix *
    q21 = nq.mult(q2, q1)
    assert_array_almost_equal, M2 @ M1, nq.quat2mat(q21)


@pytest.mark.parametrize(('M', 'q'), eg_pairs)
def test_inverse(M, q):
    iq = nq.inverse(q)
    iqM = nq.quat2mat(iq)
    iM = np.linalg.inv(M)
    assert np.allclose(iM, iqM)


@pytest.mark.parametrize('vec', np.eye(3))
@pytest.mark.parametrize(('M', 'q'), eg_pairs)
def test_qrotate(vec, M, q):
    vdash = nq.rotate_vector(vec, q)
    vM = M @ vec
    assert_array_almost_equal(vdash, vM)


@pytest.mark.parametrize('q', unit_quats)
def test_quaternion_reconstruction(q):
    # Test reconstruction of arbitrary unit quaternions
    M = nq.quat2mat(q)
    qt = nq.mat2quat(M)
    # Accept positive or negative match
    posm = np.allclose(q, qt)
    negm = np.allclose(q, -qt)
    assert posm or negm


def test_angle_axis2quat():
    q = nq.angle_axis2quat(0, [1, 0, 0])
    assert_array_equal(q, [1, 0, 0, 0])
    q = nq.angle_axis2quat(np.pi, [1, 0, 0])
    assert_array_almost_equal(q, [0, 1, 0, 0])
    q = nq.angle_axis2quat(np.pi, [1, 0, 0], True)
    assert_array_almost_equal(q, [0, 1, 0, 0])
    q = nq.angle_axis2quat(np.pi, [2, 0, 0], False)
    assert_array_almost_equal(q, [0, 1, 0, 0])


def test_angle_axis():
    for M, q in eg_pairs:
        theta, vec = nq.quat2angle_axis(q)
        q2 = nq.angle_axis2quat(theta, vec)
        nq.nearly_equivalent(q, q2)
        aa_mat = nq.angle_axis2mat(theta, vec)
        assert_array_almost_equal(aa_mat, M)
        unit_vec = norm(vec)
        aa_mat2 = nq.angle_axis2mat(theta, unit_vec, is_normalized=True)
        assert_array_almost_equal(aa_mat2, M)



import os
import tempfile

import numpy as np

import nibabel as nib
from nibabel.cifti2 import cifti2, cifti2_axes
from nibabel.tests.nibabel_data import get_nibabel_data, needs_nibabel_data

test_directory = os.path.join(get_nibabel_data(), 'nitest-cifti2')

hcp_labels = [
    'CortexLeft',
    'CortexRight',
    'AccumbensLeft',
    'AccumbensRight',
    'AmygdalaLeft',
    'AmygdalaRight',
    'brain_stem',
    'CaudateLeft',
    'CaudateRight',
    'CerebellumLeft',
    'CerebellumRight',
    'Diencephalon_ventral_left',
    'Diencephalon_ventral_right',
    'HippocampusLeft',
    'HippocampusRight',
    'PallidumLeft',
    'PallidumRight',
    'PutamenLeft',
    'PutamenRight',
    'ThalamusLeft',
    'ThalamusRight',
]

hcp_n_elements = [
    29696,
    29716,
    135,
    140,
    315,
    332,
    3472,
    728,
    755,
    8709,
    9144,
    706,
    712,
    764,
    795,
    297,
    260,
    1060,
    1010,
    1288,
    1248,
]

hcp_affine = np.array(
    [[-2.0, 0.0, 0.0, 90.0], [0.0, 2.0, 0.0, -126.0], [0.0, 0.0, 2.0, -72.0], [0.0, 0.0, 0.0, 1.0]]
)


def check_hcp_grayordinates(brain_model):
    """Checks that a BrainModelAxis matches the expected 32k HCP grayordinates"""
    assert isinstance(brain_model, cifti2_axes.BrainModelAxis)
    structures = list(brain_model.iter_structures())
    assert len(structures) == len(hcp_labels)
    idx_start = 0
    for idx, (name, _, bm), label, nel in zip(
        range(len(structures)), structures, hcp_labels, hcp_n_elements
    ):
        if idx < 2:
            assert name in bm.nvertices.keys()
            assert (bm.voxel == -1).all()
            assert (bm.vertex != -1).any()
            assert bm.nvertices[name] == 32492
        else:
            assert name not in bm.nvertices.keys()
            assert (bm.voxel != -1).any()
            assert (bm.vertex == -1).all()
            assert (bm.affine == hcp_affine).all()
            assert bm.volume_shape == (91, 109, 91)
        assert name == cifti2_axes.BrainModelAxis.to_cifti_brain_structure_name(label)
        assert len(bm) == nel
        assert (bm.name == brain_model.name[idx_start : idx_start + nel]).all()
        assert (bm.voxel == brain_model.voxel[idx_start : idx_start + nel]).all()
        assert (bm.vertex == brain_model.vertex[idx_start : idx_start + nel]).all()
        idx_start += nel
    assert idx_start == len(brain_model)

    assert (brain_model.vertex[:5] == np.arange(5)).all()
    assert structures[0][2].vertex[-1] == 32491
    assert structures[1][2].vertex[0] == 0
    assert structures[1][2].vertex[-1] == 32491
    assert structures[-1][2].name[-1] == brain_model.name[-1]
    assert (structures[-1][2].voxel[-1] == brain_model.voxel[-1]).all()
    assert structures[-1][2].vertex[-1] == brain_model.vertex[-1]
    assert (brain_model.voxel[-1] == [38, 55, 46]).all()
    assert (brain_model.voxel[70000] == [56, 22, 19]).all()


def check_Conte69(brain_model):
    """Checks that the BrainModelAxis matches the expected Conte69 surface coordinates"""
    assert isinstance(brain_model, cifti2_axes.BrainModelAxis)
    structures = list(brain_model.iter_structures())
    assert len(structures) == 2
    assert structures[0][0] == 'CIFTI_STRUCTURE_CORTEX_LEFT'
    assert structures[0][2].surface_mask.all()
    assert structures[1][0] == 'CIFTI_STRUCTURE_CORTEX_RIGHT'
    assert structures[1][2].surface_mask.all()
    assert (brain_model.voxel == -1).all()

    assert (brain_model.vertex[:5] == np.arange(5)).all()
    assert structures[0][2].vertex[-1] == 32491
    assert structures[1][2].vertex[0] == 0
    assert structures[1][2].vertex[-1] == 32491


def check_rewrite(arr, axes, extension='.nii'):
    """
    Checks whether writing the Cifti2 array to disc and reading it back in gives the same object

    Parameters
    ----------
    arr : array
        N-dimensional array of data
    axes : Sequence[cifti2_axes.Axis]
        sequence of length N with the meaning of the rows/columns along each dimension
    extension : str
        custom extension to use
    """
    (fd, name) = tempfile.mkstemp(extension)
    cifti2.Cifti2Image(arr, header=axes).to_filename(name)
    img = nib.load(name)
    arr2 = img.get_fdata()
    assert np.allclose(arr, arr2)
    for idx in range(len(img.shape)):
        assert axes[idx] == img.header.get_axis(idx)
    return img


@needs_nibabel_data('nitest-cifti2')
def test_read_ones():
    img = nib.load(os.path.join(test_directory, 'ones.dscalar.nii'))
    arr = img.get_fdata()
    axes = [img.header.get_axis(dim) for dim in range(2)]
    assert (arr == 1).all()
    assert isinstance(axes[0], cifti2_axes.ScalarAxis)
    assert len(axes[0]) == 1
    assert axes[0].name[0] == 'ones'
    assert axes[0].meta[0] == {}
    check_hcp_grayordinates(axes[1])
    img = check_rewrite(arr, axes)
    check_hcp_grayordinates(img.header.get_axis(1))


@needs_nibabel_data('nitest-cifti2')
def test_read_conte69_dscalar():
    img = nib.load(
        os.path.join(test_directory, 'Conte69.MyelinAndCorrThickness.32k_fs_LR.dscalar.nii')
    )
    arr = img.get_fdata()
    axes = [img.header.get_axis(dim) for dim in range(2)]
    assert isinstance(axes[0], cifti2_axes.ScalarAxis)
    assert len(axes[0]) == 2
    assert axes[0].name[0] == 'MyelinMap_BC_decurv'
    assert axes[0].name[1] == 'corrThickness'
    assert axes[0].meta[0] == {
        'PaletteColorMapping': '<PaletteColorMapping Version="1">\n   <ScaleMode>MODE_AUTO_SCALE_PERCENTAGE</ScaleMode>\n   <AutoScalePercentageValues>98.000000 2.000000 2.000000 98.000000</AutoScalePercentageValues>\n   <UserScaleValues>-100.000000 0.000000 0.000000 100.000000</UserScaleValues>\n   <PaletteName>ROY-BIG-BL</PaletteName>\n   <InterpolatePalette>true</InterpolatePalette>\n   <DisplayPositiveData>true</DisplayPositiveData>\n   <DisplayZeroData>false</DisplayZeroData>\n   <DisplayNegativeData>true</DisplayNegativeData>\n   <ThresholdTest>THRESHOLD_TEST_SHOW_OUTSIDE</ThresholdTest>\n   <ThresholdType>THRESHOLD_TYPE_OFF</ThresholdType>\n   <ThresholdFailureInGreen>false</ThresholdFailureInGreen>\n   <ThresholdNormalValues>-1.000000 1.000000</ThresholdNormalValues>\n   <ThresholdMappedValues>-1.000000 1.000000</ThresholdMappedValues>\n   <ThresholdMappedAvgAreaValues>-1.000000 1.000000</ThresholdMappedAvgAreaValues>\n   <ThresholdDataName></ThresholdDataName>\n   <ThresholdRangeMode>PALETTE_THRESHOLD_RANGE_MODE_MAP</ThresholdRangeMode>\n</PaletteColorMapping>'
    }
    check_Conte69(axes[1])
    check_rewrite(arr, axes)


@needs_nibabel_data('nitest-cifti2')
def test_read_conte69_dtseries():
    img = nib.load(
        os.path.join(test_directory, 'Conte69.MyelinAndCorrThickness.32k_fs_LR.dtseries.nii')
    )
    arr = img.get_fdata()
    axes = [img.header.get_axis(dim) for dim in range(2)]
    assert isinstance(axes[0], cifti2_axes.SeriesAxis)
    assert len(axes[0]) == 2
    assert axes[0].start == 0
    assert axes[0].step == 1
    assert axes[0].size == arr.shape[0]
    assert (axes[0].time == [0, 1]).all()
    check_Conte69(axes[1])
    check_rewrite(arr, axes)


@needs_nibabel_data('nitest-cifti2')
def test_read_conte69_dlabel():
    img = nib.load(
        os.path.join(test_directory, 'Conte69.parcellations_VGD11b.32k_fs_LR.dlabel.nii')
    )
    arr = img.get_fdata()
    axes = [img.header.get_axis(dim) for dim in range(2)]
    assert isinstance(axes[0], cifti2_axes.LabelAxis)
    assert len(axes[0]) == 3
    assert (
        axes[0].name
        == [
            'Composite Parcellation-lh (FRB08_OFP03_retinotopic)',
            'Brodmann lh (from colin.R via pals_R-to-fs_LR)',
            'MEDIAL WALL lh (fs_LR)',
        ]
    ).all()
    assert axes[0].label[1][70] == ('19_B05', (1.0, 0.867, 0.467, 1.0))
    assert (axes[0].meta == [{}] * 3).all()
    check_Conte69(axes[1])
    check_rewrite(arr, axes)


@needs_nibabel_data('nitest-cifti2')
def test_read_conte69_ptseries():
    img = nib.load(
        os.path.join(test_directory, 'Conte69.MyelinAndCorrThickness.32k_fs_LR.ptseries.nii')
    )
    arr = img.get_fdata()
    axes = [img.header.get_axis(dim) for dim in range(2)]
    assert isinstance(axes[0], cifti2_axes.SeriesAxis)
    assert len(axes[0]) == 2
    assert axes[0].start == 0
    assert axes[0].step == 1
    assert axes[0].size == arr.shape[0]
    assert (axes[0].time == [0, 1]).all()

    assert len(axes[1]) == 54
    voxels, vertices = axes[1]['ER_FRB08']
    assert voxels.shape == (0, 3)
    assert len(vertices) == 2
    assert vertices['CIFTI_STRUCTURE_CORTEX_LEFT'].shape == (206 // 2,)
    assert vertices['CIFTI_STRUCTURE_CORTEX_RIGHT'].shape == (206 // 2,)
    check_rewrite(arr, axes)



import getpass
import hashlib
import os
import struct
import time
import unittest
from os.path import isdir
from os.path import join as pjoin
from pathlib import Path

import numpy as np
import pytest
from numpy.testing import assert_allclose

from ...fileslice import strided_scalar
from ...testing import clear_and_catch_warnings
from ...tests.nibabel_data import get_nibabel_data, needs_nibabel_data
from ...tmpdirs import InTemporaryDirectory
from .. import (
    read_annot,
    read_geometry,
    read_label,
    read_morph_data,
    write_annot,
    write_geometry,
    write_morph_data,
)
from ..io import _pack_rgb

DATA_SDIR = 'fsaverage'

have_freesurfer = False
if 'SUBJECTS_DIR' in os.environ:
    # May have Freesurfer installed with data
    data_path = pjoin(os.environ['SUBJECTS_DIR'], DATA_SDIR)
    have_freesurfer = isdir(data_path)
else:
    # May have nibabel test data submodule checked out
    nib_data = get_nibabel_data()
    if nib_data != '':
        data_path = pjoin(nib_data, 'nitest-freesurfer', DATA_SDIR)
        have_freesurfer = isdir(data_path)

freesurfer_test = unittest.skipUnless(
    have_freesurfer, f'cannot find freesurfer {DATA_SDIR} directory'
)


@freesurfer_test
def test_geometry():
    """Test IO of .surf"""
    surf_path = pjoin(data_path, 'surf', 'lh.inflated')
    coords, faces = read_geometry(surf_path)
    assert 0 == faces.min()
    assert coords.shape[0] == faces.max() + 1

    surf_path = pjoin(data_path, 'surf', 'lh.sphere')
    coords, faces, volume_info, create_stamp = read_geometry(
        surf_path, read_metadata=True, read_stamp=True
    )

    assert 0 == faces.min()
    assert coords.shape[0] == faces.max() + 1
    assert 9 == len(volume_info)
    assert np.array_equal([2, 0, 20], volume_info['head'])
    assert create_stamp == 'created by greve on Thu Jun  8 19:17:51 2006'

    # Test equivalence of freesurfer- and nibabel-generated triangular files
    # with respect to read_geometry()
    with InTemporaryDirectory():
        surf_path = 'test'
        create_stamp = f'created by {getpass.getuser()} on {time.ctime()}'
        volume_info['cras'] = [1.0, 2.0, 3.0]
        write_geometry(surf_path, coords, faces, create_stamp, volume_info)

        coords2, faces2, volume_info2 = read_geometry(surf_path, read_metadata=True)

        for key in ('xras', 'yras', 'zras', 'cras'):
            assert_allclose(volume_info2[key], volume_info[key], rtol=1e-7, atol=1e-30)

        assert np.array_equal(volume_info2['cras'], volume_info['cras'])
        with open(surf_path, 'rb') as fobj:
            np.fromfile(fobj, '>u1', 3)
            read_create_stamp = fobj.readline().decode().rstrip('\n')

        # now write an incomplete file
        write_geometry(surf_path, coords, faces)
        with pytest.warns(UserWarning) as w:
            read_geometry(surf_path, read_metadata=True)
        assert any('volume information contained' in str(ww.message) for ww in w)
        assert any('extension code' in str(ww.message) for ww in w)

        volume_info['head'] = [1, 2]
        with pytest.warns(UserWarning, match='Unknown extension'):
            write_geometry(surf_path, coords, faces, create_stamp, volume_info)

        volume_info['a'] = 0
        with pytest.raises(ValueError):
            write_geometry(surf_path, coords, faces, create_stamp, volume_info)

    assert create_stamp == read_create_stamp

    assert np.array_equal(coords, coords2)
    assert np.array_equal(faces, faces2)

    # Validate byte ordering
    coords_swapped = coords.byteswap()
    coords_swapped = coords_swapped.view(coords_swapped.dtype.newbyteorder())
    faces_swapped = faces.byteswap()
    faces_swapped = faces_swapped.view(faces_swapped.dtype.newbyteorder())
    assert np.array_equal(coords_swapped, coords)
    assert np.array_equal(faces_swapped, faces)


@freesurfer_test
@needs_nibabel_data('nitest-freesurfer')
def test_quad_geometry():
    """Test IO of freesurfer quad files."""
    new_quad = pjoin(
        get_nibabel_data(), 'nitest-freesurfer', 'subjects', 'bert', 'surf', 'lh.inflated.nofix'
    )
    coords, faces = read_geometry(new_quad)
    assert 0 == faces.min()
    assert coords.shape[0] == (faces.max() + 1)
    with InTemporaryDirectory():
        new_path = 'test'
        write_geometry(new_path, coords, faces)
        coords2, faces2 = read_geometry(new_path)
        assert np.array_equal(coords, coords2)
        assert np.array_equal(faces, faces2)


@freesurfer_test
def test_morph_data():
    """Test IO of morphometry data file (eg. curvature)."""
    curv_path = pjoin(data_path, 'surf', 'lh.curv')
    curv = read_morph_data(curv_path)
    assert -1.0 < curv.min() < 0
    assert 0 < curv.max() < 1.0
    with InTemporaryDirectory():
        new_path = 'test'
        write_morph_data(new_path, curv)
        curv2 = read_morph_data(new_path)
        assert np.array_equal(curv2, curv)


def test_write_morph_data():
    """Test write_morph_data edge cases"""
    values = np.arange(20, dtype='>f4')
    okay_shapes = [(20,), (20, 1), (20, 1, 1), (1, 20)]
    bad_shapes = [(10, 2), (1, 1, 20, 1, 1)]
    big_num = np.iinfo('i4').max + 1
    with InTemporaryDirectory():
        for shape in okay_shapes:
            write_morph_data('test.curv', values.reshape(shape))
            # Check ordering is preserved, regardless of shape
            assert np.array_equal(read_morph_data('test.curv'), values)

        with pytest.raises(ValueError):
            write_morph_data('test.curv', np.zeros(shape), big_num)
        # Windows 32-bit overflows Python int
        if np.dtype(int) != np.dtype(np.int32):
            with pytest.raises(ValueError):
                write_morph_data('test.curv', strided_scalar((big_num,)))
        for shape in bad_shapes:
            with pytest.raises(ValueError):
                write_morph_data('test.curv', values.reshape(shape))


@freesurfer_test
def test_annot():
    """Test IO of .annot against freesurfer example data."""
    annots = ['aparc', 'aparc.a2005s']
    for a in annots:
        annot_path = pjoin(data_path, 'label', f'lh.{a}.annot')

        labels, ctab, names = read_annot(annot_path)
        assert labels.shape == (163842,)
        assert ctab.shape == (len(names), 5)

        labels_orig = None
        if a == 'aparc':
            labels_orig, _, _ = read_annot(annot_path, orig_ids=True)
            np.testing.assert_array_equal(labels == -1, labels_orig == 0)
            # Handle different version of fsaverage
            content_hash = hashlib.md5(Path(annot_path).read_bytes()).hexdigest()
            if content_hash == 'bf0b488994657435cdddac5f107d21e8':
                assert np.sum(labels_orig == 0) == 13887
            elif content_hash == 'd4f5b7cbc2ed363ac6fcf89e19353504':
                assert np.sum(labels_orig == 1639705) == 13327
            else:
                raise RuntimeError(
                    'Unknown freesurfer file. Please report '
                    'the problem to the maintainer of nibabel.'
                )

        # Test equivalence of freesurfer- and nibabel-generated annot files
        # with respect to read_annot()
        with InTemporaryDirectory():
            annot_path = 'test'
            write_annot(annot_path, labels, ctab, names)

            labels2, ctab2, names2 = read_annot(annot_path)
            if labels_orig is not None:
                labels_orig_2, _, _ = read_annot(annot_path, orig_ids=True)

        assert np.array_equal(labels, labels2)
        if labels_orig is not None:
            assert np.array_equal(labels_orig, labels_orig_2)
        assert np.array_equal(ctab, ctab2)
        assert names == names2


def test_read_write_annot():
    """Test generating .annot file and reading it back."""
    # This annot file will store a LUT for a mesh made of 10 vertices, with
    # 3 colours in the LUT.
    nvertices = 10
    nlabels = 3
    names = [f'label {l}' for l in range(1, nlabels + 1)]
    # randomly generate a label for each vertex, making sure
    # that at least one of each label value is present. Label
    # values are in the range (0, nlabels-1) - they are used
    # as indices into the lookup table (generated below).
    labels = list(range(nlabels)) + list(np.random.randint(0, nlabels, nvertices - nlabels))
    labels = np.array(labels, dtype=np.int32)
    np.random.shuffle(labels)
    # Generate some random colours for the LUT
    rgbal = np.zeros((nlabels, 5), dtype=np.int32)
    rgbal[:, :4] = np.random.randint(0, 255, (nlabels, 4))
    # But make sure we have at least one large alpha, to make sure that when
    # it is packed into a signed 32 bit int, it results in a negative value
    # for the annotation value.
    rgbal[0, 3] = 255
    # Generate the annotation values for each LUT entry
    rgbal[:, 4] = rgbal[:, 0] + rgbal[:, 1] * (2**8) + rgbal[:, 2] * (2**16)
    annot_path = 'c.annot'
    with InTemporaryDirectory():
        write_annot(annot_path, labels, rgbal, names, fill_ctab=False)
        labels2, rgbal2, names2 = read_annot(annot_path)
        names2 = [n.decode('ascii') for n in names2]
        assert np.all(np.isclose(rgbal2, rgbal))
        assert np.all(np.isclose(labels2, labels))
        assert names2 == names


def test_write_annot_fill_ctab():
    """Test the `fill_ctab` parameter to :func:`.write_annot`."""
    nvertices = 10
    nlabels = 3
    names = [f'label {l}' for l in range(1, nlabels + 1)]
    labels = list(range(nlabels)) + list(np.random.randint(0, nlabels, nvertices - nlabels))
    labels = np.array(labels, dtype=np.int32)
    np.random.shuffle(labels)
    rgba = np.array(np.random.randint(0, 255, (nlabels, 4)), dtype=np.int32)
    annot_path = 'c.annot'
    with InTemporaryDirectory():
        write_annot(annot_path, labels, rgba, names, fill_ctab=True)
        labels2, rgbal2, names2 = read_annot(annot_path)
        names2 = [n.decode('ascii') for n in names2]
        assert np.all(np.isclose(rgbal2[:, :4], rgba))
        assert np.all(np.isclose(labels2, labels))
        assert names2 == names
        # make sure a warning is emitted if fill_ctab is False, and the
        # annotation values are wrong. Use orig_ids=True so we get those bad
        # values back.
        badannot = (10 * np.arange(nlabels, dtype=np.int32)).reshape(-1, 1)
        rgbal = np.hstack((rgba, badannot))
        with pytest.warns(
            UserWarning, match=f'Annotation values in {annot_path} will be incorrect'
        ):
            write_annot(annot_path, labels, rgbal, names, fill_ctab=False)
        labels2, rgbal2, names2 = read_annot(annot_path, orig_ids=True)
        names2 = [n.decode('ascii') for n in names2]
        assert np.all(np.isclose(rgbal2[:, :4], rgba))
        assert np.all(np.isclose(labels2, badannot[labels].squeeze()))
        assert names2 == names
        # make sure a warning is *not* emitted if fill_ctab is False, but the
        # annotation values are correct.
        rgbal = np.hstack((rgba, np.zeros((nlabels, 1), dtype=np.int32)))
        rgbal[:, 4] = rgbal[:, 0] + rgbal[:, 1] * (2**8) + rgbal[:, 2] * (2**16)
        with clear_and_catch_warnings() as w:
            write_annot(annot_path, labels, rgbal, names, fill_ctab=False)
        assert all(
            f'Annotation values in {annot_path} will be incorrect' != str(ww.message) for ww in w
        )
        labels2, rgbal2, names2 = read_annot(annot_path)
        names2 = [n.decode('ascii') for n in names2]
        assert np.all(np.isclose(rgbal2[:, :4], rgba))
        assert np.all(np.isclose(labels2, labels))
        assert names2 == names


def test_read_annot_old_format():
    """Test reading an old-style .annot file."""

    def gen_old_annot_file(fpath, nverts, labels, rgba, names):
        dt = '>i'
        vdata = np.zeros((nverts, 2), dtype=dt)
        vdata[:, 0] = np.arange(nverts)
        vdata[:, [1]] = _pack_rgb(rgba[labels, :3])
        fbytes = b''
        # number of vertices
        fbytes += struct.pack(dt, nverts)
        # vertices + annotation values
        fbytes += vdata.astype(dt).tobytes()
        # is there a colour table?
        fbytes += struct.pack(dt, 1)
        # number of entries in colour table
        fbytes += struct.pack(dt, rgba.shape[0])
        # length of orig_tab string
        fbytes += struct.pack(dt, 5)
        fbytes += b'abcd\x00'
        for i in range(rgba.shape[0]):
            # length of entry name (+1 for terminating byte)
            fbytes += struct.pack(dt, len(names[i]) + 1)
            fbytes += names[i].encode('ascii') + b'\x00'
            fbytes += rgba[i, :].astype(dt).tobytes()
        with open(fpath, 'wb') as f:
            f.write(fbytes)

    with InTemporaryDirectory():
        nverts = 10
        nlabels = 3
        names = [f'Label {l}' for l in range(nlabels)]
        labels = np.concatenate(
            (np.arange(nlabels), np.random.randint(0, nlabels, nverts - nlabels))
        )
        np.random.shuffle(labels)
        rgba = np.random.randint(0, 255, (nlabels, 4))
        # write an old .annot file
        gen_old_annot_file('blah.annot', nverts, labels, rgba, names)
        # read it back
        rlabels, rrgba, rnames = read_annot('blah.annot')
        rnames = [n.decode('ascii') for n in rnames]
        assert np.all(np.isclose(labels, rlabels))
        assert np.all(np.isclose(rgba, rrgba[:, :4]))
        assert names == rnames


@freesurfer_test
def test_label():
    """Test IO of .label"""
    label_path = pjoin(data_path, 'label', 'lh.cortex.label')
    label = read_label(label_path)
    # XXX : test more
    assert label.min() >= 0
    assert label.max() <= 163841
    assert label.shape[0] <= 163842

    labels, scalars = read_label(label_path, True)
    assert np.all(labels == label)
    assert len(labels) == len(scalars)


def test_write_annot_maxstruct():
    """Test writing ANNOT files with repeated labels"""
    with InTemporaryDirectory():
        nlabels = 3
        names = [f'label {l}' for l in range(1, nlabels + 1)]
        # max label < n_labels
        labels = np.array([1, 1, 1], dtype=np.int32)
        rgba = np.array(np.random.randint(0, 255, (nlabels, 4)), dtype=np.int32)
        annot_path = 'c.annot'

        write_annot(annot_path, labels, rgba, names)
        # Validate the file can be read
        rt_labels, rt_ctab, rt_names = read_annot(annot_path)
        # Check round-trip
        assert np.array_equal(labels, rt_labels)
        assert np.array_equal(rgba, rt_ctab[:, :4])
        assert names == [n.decode('ascii') for n in rt_names]



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import unittest
from collections import namedtuple as nt

import numpy as np
import pytest
from numpy.testing import assert_array_equal, assert_equal

from ..optpkg import optional_package
from ..viewers import OrthoSlicer3D

# Need at least MPL 1.3 for viewer tests.
# 2020.02.11 - 1.3 wheels are no longer distributed, so the minimum we test with is 1.5
matplotlib, has_mpl, _ = optional_package('matplotlib', min_version='1.5')

needs_mpl = unittest.skipUnless(has_mpl, 'These tests need matplotlib')
if has_mpl:
    matplotlib.use('Agg')


@needs_mpl
def test_viewer():
    # Test viewer
    plt = optional_package('matplotlib.pyplot')[0]
    a = np.sin(np.linspace(0, np.pi, 20))
    b = np.sin(np.linspace(0, np.pi * 5, 30))
    data = (np.outer(a, b)[..., np.newaxis] * a)[:, :, :, np.newaxis]
    data = data * np.array([1.0, 2.0])  # give it a # of volumes > 1
    v = OrthoSlicer3D(data)
    assert_array_equal(v.position, (0, 0, 0))
    assert 'OrthoSlicer3D' in repr(v)

    # fake some events, inside and outside axes
    v._on_scroll(nt('event', 'button inaxes key')('up', None, None))
    for ax in (v._axes[0], v._axes[3]):
        v._on_scroll(nt('event', 'button inaxes key')('up', ax, None))
    v._on_scroll(nt('event', 'button inaxes key')('up', ax, 'shift'))
    # "click" outside axes, then once in each axis, then move without click
    v._on_mouse(nt('event', 'xdata ydata inaxes button')(0.5, 0.5, None, 1))
    for ax in v._axes:
        v._on_mouse(nt('event', 'xdata ydata inaxes button')(0.5, 0.5, ax, 1))
    v._on_mouse(nt('event', 'xdata ydata inaxes button')(0.5, 0.5, None, None))
    v.set_volume_idx(1)
    v.cmap = 'hot'
    v.clim = (0, 3)
    with pytest.raises(ValueError):
        OrthoSlicer3D.clim.fset(v, (0.0,))  # bad limits
    with pytest.raises(
        (
            ValueError,  # MPL3.5 and lower
            KeyError,  # MPL3.6 and higher
        )
    ):
        OrthoSlicer3D.cmap.fset(v, 'foo')  # wrong cmap

    # decrement/increment volume numbers via keypress
    v.set_volume_idx(1)  # should just pass
    v._on_keypress(nt('event', 'key')('-'))  # decrement
    assert_equal(v._data_idx[3], 0)
    v._on_keypress(nt('event', 'key')('+'))  # increment
    assert_equal(v._data_idx[3], 1)
    v._on_keypress(nt('event', 'key')('-'))
    v._on_keypress(nt('event', 'key')('='))  # alternative increment key
    assert_equal(v._data_idx[3], 1)

    v.close()
    v._draw()  # should be safe

    # non-multi-volume
    v = OrthoSlicer3D(data[:, :, :, 0])
    v._on_scroll(nt('event', 'button inaxes key')('up', v._axes[0], 'shift'))
    v._on_keypress(nt('event', 'key')('escape'))
    v.close()

    # complex input should raise a TypeError prior to figure creation
    with pytest.raises(TypeError):
        OrthoSlicer3D(data[:, :, :, 0].astype(np.complex64))

    # other cases
    fig, axes = plt.subplots(1, 4)
    v1 = OrthoSlicer3D(data, axes=axes)
    aff = np.array([[0, 1, 0, 3], [-1, 0, 0, 2], [0, 0, 2, 1], [0, 0, 0, 1]], float)
    v2 = OrthoSlicer3D(data, affine=aff, axes=axes[:3])
    # bad data (not 3+ dim)
    with pytest.raises(ValueError):
        OrthoSlicer3D(data[:, :, 0, 0])
    # bad affine (not 4x4)
    with pytest.raises(ValueError):
        OrthoSlicer3D(data, affine=np.eye(3))
    with pytest.raises(TypeError):
        v2.link_to(1)
    v2.link_to(v1)
    v2.link_to(v1)  # shouldn't do anything
    v1.close()
    v2.close()
    plt.close(fig)


@needs_mpl
def test_viewer_nonRAS():
    data1 = np.random.rand(10, 20, 40)
    data1[5, 10, :] = 0
    data1[5, :, 30] = 0
    data1[:, 10, 30] = 0
    # RSA affine
    aff1 = np.array([[1, 0, 0, -5], [0, 0, 1, -30], [0, 1, 0, -10], [0, 0, 0, 1]])
    o1 = OrthoSlicer3D(data1, aff1)
    sag = o1._ims[0].get_array()
    cor = o1._ims[1].get_array()
    axi = o1._ims[2].get_array()

    # Sagittal view: [0, I->S, P->A], so data is transposed, matching plot array
    assert_array_equal(sag, data1[5, :, :])
    # Coronal view: [L->R, I->S, 0]. Data is not transposed, transpose to match plot array
    assert_array_equal(cor, data1[:, :, 30].T)
    # Axial view: [L->R, 0, P->A]. Data is not transposed, transpose to match plot array
    assert_array_equal(axi, data1[:, 10, :].T)

    o1.set_position(1, 2, 3)  # R, A, S coordinates

    sag = o1._ims[0].get_array()
    cor = o1._ims[1].get_array()
    axi = o1._ims[2].get_array()

    # Shift 1 right, 2 anterior, 3 superior
    assert_array_equal(sag, data1[6, :, :])
    assert_array_equal(cor, data1[:, :, 32].T)
    assert_array_equal(axi, data1[:, 13, :].T)


@needs_mpl
def test_viewer_nonRAS_on_mouse():
    """
    test on_mouse selection on non RAS matrices

    """
    # This affine simulates an acquisition on a quadruped subject that is in a prone position.
    # This corresponds to an acquisition with:
    # - LR inverted on scanner x (i)
    # - IS on scanner y (j)
    # - PA on scanner z (k)
    # This example enables to test also OrthoSlicer3D properties `_flips` and `_order`.

    (I, J, K) = (10, 20, 40)
    data1 = np.random.rand(I, J, K)
    (i_target, j_target, k_target) = (2, 14, 12)
    i1 = i_target - 2
    i2 = i_target + 2
    j1 = j_target - 3
    j2 = j_target + 3
    k1 = k_target - 4
    k2 = k_target + 4
    data1[i1 : i2 + 1, j1 : j2 + 1, k1 : k2 + 1] = 0
    data1[i_target, j_target, k_target] = 1
    valp1 = 1.5
    valm1 = 0.5
    data1[i_target - 1, j_target, k_target] = valp1  # x flipped
    data1[i_target + 1, j_target, k_target] = valm1  # x flipped
    data1[i_target, j_target - 1, k_target] = valm1
    data1[i_target, j_target + 1, k_target] = valp1
    data1[i_target, j_target, k_target - 1] = valm1
    data1[i_target, j_target, k_target + 1] = valp1

    aff1 = np.array([[-1, 0, 0, 5], [0, 0, 1, -10], [0, 1, 0, -30], [0, 0, 0, 1]])

    o1 = OrthoSlicer3D(data1, aff1)

    class Event:
        def __init__(self):
            self.name = 'simulated mouse event'
            self.button = 1

    event = Event()
    event.xdata = k_target
    event.ydata = j_target
    event.inaxes = o1._ims[0].axes
    o1._on_mouse(event)

    event.inaxes = o1._ims[1].axes
    event.xdata = (I - 1) - i_target  # x flipped
    event.ydata = j_target
    o1._on_mouse(event)

    event.inaxes = o1._ims[2].axes
    event.xdata = (I - 1) - i_target  # x flipped
    event.ydata = k_target
    o1._on_mouse(event)

    sag = o1._ims[0].get_array()
    cor = o1._ims[1].get_array()
    axi = o1._ims[2].get_array()

    assert_array_equal(sag, data1[i_target, :, :])  #
    assert_array_equal(cor, data1[::-1, :, k_target].T)  # x flipped
    assert_array_equal(axi, data1[::-1, j_target, :].T)  # x flipped
    return None


@needs_mpl
def test_viewer_nonRAS_on_scroll():
    """
    test scrolling on non RAS matrices

    """
    # This affine simulates an acquisition on a quadruped subject that is in a prone position.
    # This corresponds to an acquisition with:
    # - LR inverted on scanner x (i)
    # - IS on scanner y (j)
    # - PA on scanner z (k)
    # This example enables to test also OrthoSlicer3D properties `_flips` and `_order`.

    (I, J, K) = (10, 20, 40)
    data1 = np.random.rand(I, J, K)
    (i_target, j_target, k_target) = (2, 14, 12)
    i1 = i_target - 2
    i2 = i_target + 2
    j1 = j_target - 3
    j2 = j_target + 3
    k1 = k_target - 4
    k2 = k_target + 4
    data1[i1 : i2 + 1, j1 : j2 + 1, k1 : k2 + 1] = 0
    data1[i_target, j_target, k_target] = 1
    valp1 = 1.5
    valm1 = 0.5
    data1[i_target - 1, j_target, k_target] = valp1  # x flipped
    data1[i_target + 1, j_target, k_target] = valm1  # x flipped
    data1[i_target, j_target - 1, k_target] = valm1
    data1[i_target, j_target + 1, k_target] = valp1
    data1[i_target, j_target, k_target - 1] = valm1
    data1[i_target, j_target, k_target + 1] = valp1

    aff1 = np.array([[-1, 0, 0, 5], [0, 0, 1, -10], [0, 1, 0, -30], [0, 0, 0, 1]])

    o1 = OrthoSlicer3D(data1, aff1)

    class Event:
        def __init__(self):
            self.name = 'simulated mouse event'
            self.button = None
            self.key = None

    [x_t, y_t, z_t] = list(aff1.dot(np.array([i_target, j_target, k_target, 1]))[:3])
    # print(x_t, y_t, z_t)
    # scanner positions are x_t=3, y_t=2, z_t=16

    event = Event()

    # Sagittal plane - one scroll up
    # x coordinate is flipped so index decrease by 1
    o1.set_position(x_t, y_t, z_t)
    event.inaxes = o1._ims[0].axes
    event.button = 'up'
    o1._on_scroll(event)
    sag = o1._ims[0].get_array()
    cor = o1._ims[1].get_array()
    axi = o1._ims[2].get_array()
    assert_array_equal(sag, data1[i_target - 1, :, :])
    assert_array_equal(cor, data1[::-1, :, k_target].T)  # ::-1 because the array is flipped in x
    assert_array_equal(axi, data1[::-1, j_target, :].T)  # ::-1 because the array is flipped in x

    # Sagittal plane - one scrolled down
    o1.set_position(x_t, y_t, z_t)
    event.button = 'down'
    o1._on_scroll(event)
    sag = o1._ims[0].get_array()
    cor = o1._ims[1].get_array()
    axi = o1._ims[2].get_array()
    assert_array_equal(sag, data1[i_target + 1, :, :])
    assert_array_equal(cor, data1[::-1, :, k_target].T)
    assert_array_equal(axi, data1[::-1, j_target, :].T)

    # Coronal plane - one scroll up
    # y coordinate is increase by 1
    o1.set_position(x_t, y_t, z_t)
    event.inaxes = o1._ims[1].axes
    event.button = 'up'
    o1._on_scroll(event)
    sag = o1._ims[0].get_array()
    cor = o1._ims[1].get_array()
    axi = o1._ims[2].get_array()
    assert_array_equal(sag, data1[i_target, :, :])
    assert_array_equal(
        cor, data1[::-1, :, k_target + 1].T
    )  # ::-1 because the array is flipped in x
    assert_array_equal(axi, data1[::-1, j_target, :].T)  # ::-1 because the array is flipped in x

    # Coronal plane - one scrolled down
    o1.set_position(x_t, y_t, z_t)
    event.button = 'down'
    o1._on_scroll(event)
    sag = o1._ims[0].get_array()
    cor = o1._ims[1].get_array()
    axi = o1._ims[2].get_array()
    assert_array_equal(sag, data1[i_target, :, :])
    assert_array_equal(cor, data1[::-1, :, k_target - 1].T)
    assert_array_equal(axi, data1[::-1, j_target, :].T)

    # Axial plane - one scroll up
    # y is increase by 1
    o1.set_position(x_t, y_t, z_t)
    event.inaxes = o1._ims[2].axes
    event.button = 'up'
    o1._on_scroll(event)
    sag = o1._ims[0].get_array()
    cor = o1._ims[1].get_array()
    axi = o1._ims[2].get_array()
    assert_array_equal(sag, data1[i_target, :, :])
    assert_array_equal(cor, data1[::-1, :, k_target].T)  # ::-1 because the array is flipped in x
    assert_array_equal(
        axi, data1[::-1, j_target + 1, :].T
    )  # ::-1 because the array is flipped in x

    # Axial plane - one scrolled down
    o1.set_position(x_t, y_t, z_t)
    event.button = 'down'
    o1._on_scroll(event)
    sag = o1._ims[0].get_array()
    cor = o1._ims[1].get_array()
    axi = o1._ims[2].get_array()
    assert_array_equal(sag, data1[i_target, :, :])
    assert_array_equal(cor, data1[::-1, :, k_target].T)
    assert_array_equal(axi, data1[::-1, j_target - 1, :].T)
    return None



"""Utilities for working with DICOM datasets"""

from enum import Enum


def find_private_section(dcm_data, group_no, creator):
    """Return start element in group `group_no` given creator name `creator`

    Private attribute tags need to announce where they will go by putting a tag
    in the private group (here `group_no`) between elements 1 and 0xFF.  The
    element number of these tags give the start of matching information, in the
    higher tag numbers.

    Parameters
    ----------
    dcm_data : dicom ``dataset``
        Iterating over `dcm_data` produces ``elements`` with attributes
        ``tag``, ``VR``, ``value``
    group_no : int
        Group number in which to search
    creator : str or bytes or regex
        Name of section - e.g. 'SIEMENS CSA HEADER' - or regex to search for
        section name.  Regex used via ``creator.search(element_value)`` where
        ``element_value`` is the value of the data element.

    Returns
    -------
    element_start : int
        Element number at which named section starts.
    """
    if hasattr(creator, 'search'):
        match_func = creator.search
    else:
        if isinstance(creator, bytes):
            creator = creator.decode('latin-1')
        match_func = creator.__eq__
    # Group elements assumed ordered by tag (groupno, elno)
    for element in dcm_data.group_dataset(group_no):
        elno = element.tag.elem
        if elno > 0xFF:
            break
        if element.VR not in ('LO', 'OB'):
            continue
        val = element.value
        if isinstance(val, bytes):
            val = val.decode('latin-1')
        if match_func(val):
            return elno * 0x100
    return None


class Vendor(Enum):
    SIEMENS = 1
    GE = 2
    PHILIPS = 3


vendor_priv_sections = {
    Vendor.SIEMENS: [
        (0x9, 'SIEMENS SYNGO INDEX SERVICE'),
        (0x19, 'SIEMENS MR HEADER'),
        (0x21, 'SIEMENS MR SDR 01'),
        (0x21, 'SIEMENS MR SDS 01'),
        (0x21, 'SIEMENS MR SDI 02'),
        (0x29, 'SIEMENS CSA HEADER'),
        (0x29, 'SIEMENS MEDCOM HEADER2'),
        (0x51, 'SIEMENS MR HEADER'),
    ],
    Vendor.PHILIPS: [
        (0x2001, 'Philips Imaging DD 001'),
        (0x2001, 'Philips Imaging DD 002'),
        (0x2001, 'Philips Imaging DD 129'),
        (0x2005, 'Philips MR Imaging DD 001'),
        (0x2005, 'Philips MR Imaging DD 002'),
        (0x2005, 'Philips MR Imaging DD 003'),
        (0x2005, 'Philips MR Imaging DD 004'),
        (0x2005, 'Philips MR Imaging DD 005'),
        (0x2005, 'Philips MR Imaging DD 006'),
        (0x2005, 'Philips MR Imaging DD 007'),
        (0x2005, 'Philips MR Imaging DD 005'),
        (0x2005, 'Philips MR Imaging DD 006'),
    ],
    Vendor.GE: [
        (0x9, 'GEMS_IDEN_01'),
        (0x19, 'GEMS_ACQU_01'),
        (0x21, 'GEMS_RELA_01'),
        (0x23, 'GEMS_STDY_01'),
        (0x25, 'GEMS_SERS_01'),
        (0x27, 'GEMS_IMAG_01'),
        (0x29, 'GEMS_IMPS_01'),
        (0x43, 'GEMS_PARM_01'),
    ],
}


def vendor_from_private(dcm_data):
    """Try to determine the vendor by looking for specific private tags"""
    for vendor, priv_sections in vendor_priv_sections.items():
        for priv_group, priv_creator in priv_sections:
            if find_private_section(dcm_data, priv_group, priv_creator) != None:
                return vendor



"""Benchmarks for array_to_file routine

Run benchmarks with::

    import nibabel as nib
    nib.bench()

Run this benchmark with::

    pytest -c <path>/benchmarks/pytest.benchmark.ini <path>/benchmarks/bench_array_to_file.py
"""

import sys
from io import BytesIO  # noqa: F401

import numpy as np
from numpy.testing import measure

from nibabel.volumeutils import array_to_file  # noqa: F401

from .butils import print_git_title


def bench_array_to_file():
    rng = np.random.RandomState(20111001)
    repeat = 10
    img_shape = (128, 128, 64, 10)
    arr = rng.normal(size=img_shape)
    sys.stdout.flush()
    print_git_title('\nArray to file')
    mtime = measure('array_to_file(arr, BytesIO(), np.float32)', repeat)
    fmt = '{:30s} {:6.2f}'.format
    print(fmt('Save float64 to float32', mtime))
    mtime = measure('array_to_file(arr, BytesIO(), np.int16)', repeat)
    print(fmt('Save float64 to int16', mtime))
    # Set a lot of NaNs to check timing
    arr[:, :, :, 1] = np.nan
    mtime = measure('array_to_file(arr, BytesIO(), np.float32)', repeat)
    print(fmt('Save float64 to float32, NaNs', mtime))
    mtime = measure('array_to_file(arr, BytesIO(), np.int16)', repeat)
    print(fmt('Save float64 to int16, NaNs', mtime))
    # Set a lot of infs to check timing
    arr[:, :, :, 1] = np.inf
    mtime = measure('array_to_file(arr, BytesIO(), np.float32)', repeat)
    print(fmt('Save float64 to float32, infs', mtime))
    mtime = measure('array_to_file(arr, BytesIO(), np.int16)', repeat)
    print(fmt('Save float64 to int16, infs', mtime))
    # Int16 input, float output
    arr = np.random.random_integers(low=-1000, high=1000, size=img_shape)
    arr = arr.astype(np.int16)
    mtime = measure('array_to_file(arr, BytesIO(), np.float32)', repeat)
    print(fmt('Save Int16 to float32', mtime))
    sys.stdout.flush()



"""Make anatomical image with altered affine

* Add some rotations and translations to affine;
* Save as ``.nii`` file so SPM can read it.

See ``resample_using_spm.m`` for processing of this generated image by SPM.
"""

import numpy as np

import nibabel as nib
from nibabel.affines import from_matvec
from nibabel.eulerangles import euler2mat

if __name__ == '__main__':
    img = nib.load('anatomical.nii')
    some_rotations = euler2mat(0.1, 0.2, 0.3)
    extra_affine = from_matvec(some_rotations, [3, 4, 5])
    moved_anat = nib.Nifti1Image(img.dataobj, extra_affine.dot(img.affine), img.header)
    moved_anat.set_data_dtype(np.float32)
    nib.save(moved_anat, 'anat_moved.nii')



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Testing filesets - a draft"""

from io import BytesIO

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from .. import MGHImage, Nifti1Image, Nifti1Pair, all_image_classes
from ..fileholders import FileHolderError
from ..spatialimages import SpatialImage


def test_files_spatialimages():
    # test files creation in image classes
    arr = np.zeros((2, 3, 4))
    aff = np.eye(4)
    klasses = [
        klass for klass in all_image_classes if klass.rw and issubclass(klass, SpatialImage)
    ]
    for klass in klasses:
        file_map = klass.make_file_map()
        for value in file_map.values():
            assert value.filename is None
            assert value.fileobj is None
            assert value.pos == 0
        # If we can't create new images in memory without loading, bail here
        if not klass.makeable:
            continue
        # MGHImage accepts only a few datatypes
        # so we force a type change to float32
        if klass == MGHImage:
            img = klass(arr.astype(np.float32), aff)
        else:
            img = klass(arr, aff)
        for value in img.file_map.values():
            assert value.filename is None
            assert value.fileobj is None
            assert value.pos == 0


def test_files_interface():
    # test high-level interface to files mapping
    arr = np.zeros((2, 3, 4))
    aff = np.eye(4)
    img = Nifti1Image(arr, aff)
    # single image
    img.set_filename('test')
    assert img.get_filename() == 'test.nii'
    assert img.file_map['image'].filename == 'test.nii'
    with pytest.raises(KeyError):
        img.file_map['header']
    # pair - note new class
    img = Nifti1Pair(arr, aff)
    img.set_filename('test')
    assert img.get_filename() == 'test.img'
    assert img.file_map['image'].filename == 'test.img'
    assert img.file_map['header'].filename == 'test.hdr'
    # fileobjs - single image
    img = Nifti1Image(arr, aff)
    img.file_map['image'].fileobj = BytesIO()
    img.to_file_map()  # saves to files
    img2 = Nifti1Image.from_file_map(img.file_map)
    # img still has correct data
    assert_array_equal(img2.get_fdata(), img.get_fdata())
    # fileobjs - pair
    img = Nifti1Pair(arr, aff)
    img.file_map['image'].fileobj = BytesIO()
    # no header yet
    with pytest.raises(FileHolderError):
        img.to_file_map()
    img.file_map['header'].fileobj = BytesIO()
    img.to_file_map()  # saves to files
    img2 = Nifti1Pair.from_file_map(img.file_map)
    # img still has correct data
    assert_array_equal(img2.get_fdata(), img.get_fdata())


def test_round_trip_spatialimages():
    # write an image to files
    data = np.arange(24, dtype='i4').reshape((2, 3, 4))
    aff = np.eye(4)
    klasses = [
        klass
        for klass in all_image_classes
        if klass.rw and klass.makeable and issubclass(klass, SpatialImage)
    ]
    for klass in klasses:
        file_map = klass.make_file_map()
        for key in file_map:
            file_map[key].fileobj = BytesIO()
        img = klass(data, aff)
        img.file_map = file_map
        img.to_file_map()
        # read it back again from the written files
        img2 = klass.from_file_map(file_map)
        assert_array_equal(img2.get_fdata(), data)
        # write, read it again
        img2.to_file_map()
        img3 = klass.from_file_map(file_map)
        assert_array_equal(img3.get_fdata(), data)



#!python
# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Output a summary table for neuroimaging files (resolution, dimensionality, etc.)
"""

import sys
from optparse import Option, OptionParser

import numpy as np

import nibabel as nib
import nibabel.cmdline.utils
from nibabel.cmdline.utils import _err, ap, safe_get, table2string, verbose

__copyright__ = 'Copyright (c) 2011-18 Yaroslav Halchenko and NiBabel contributors'
__license__ = 'MIT'


MAX_UNIQUE = 1000  # maximal number of unique values to report for --counts


def get_opt_parser():
    # use module docstring for help output
    p = OptionParser(
        usage=f'{sys.argv[0]} [OPTIONS] [FILE ...]\n\n' + __doc__,
        version='%prog ' + nib.__version__,
    )

    p.add_options(
        [
            Option(
                '-v',
                '--verbose',
                action='count',
                dest='verbose',
                default=0,
                help='Make more noise.  Could be specified multiple times',
            ),
            Option(
                '-H',
                '--header-fields',
                dest='header_fields',
                default='',
                help='Header fields (comma separated) to be printed as well (if present)',
            ),
            Option(
                '-s',
                '--stats',
                action='store_true',
                dest='stats',
                default=False,
                help='Output basic data statistics',
            ),
            Option(
                '-c',
                '--counts',
                action='store_true',
                dest='counts',
                default=False,
                help='Output counts - number of entries for each numeric value '
                '(useful for int ROI maps)',
            ),
            Option(
                '--all-counts',
                action='store_true',
                dest='all_counts',
                default=False,
                help=f'Output all counts, even if number of unique values > {MAX_UNIQUE}',
            ),
            Option(
                '-z',
                '--zeros',
                action='store_true',
                dest='stats_zeros',
                default=False,
                help='Include zeros into output basic data statistics (--stats, --counts)',
            ),
        ]
    )

    return p


def proc_file(f, opts):
    verbose(1, f'Loading {f}')

    row = [f'@l{f}']
    try:
        vol = nib.load(f)
        h = vol.header
    except Exception as e:
        row += ['failed']
        verbose(2, f'Failed to gather information -- {e}')
        return row

    row += [
        str(safe_get(h, 'data_dtype')),
        f'@l[{ap(safe_get(h, "data_shape"), "%3g")}]',
        f'@l{ap(safe_get(h, "zooms"), "%.2f", "x")}',
    ]
    # Slope
    if (
        hasattr(h, 'has_data_slope')
        and (h.has_data_slope or h.has_data_intercept)
        and not h.get_slope_inter() in ((1.0, 0.0), (None, None))
    ):
        row += ['@l*{:.3g}+{:.3g}'.format(*h.get_slope_inter())]
    else:
        row += ['']

    if hasattr(h, 'extensions') and len(h.extensions):
        row += [f'@l#exts: {len(h.extensions)}']
    else:
        row += ['']

    if opts.header_fields:
        # signals "all fields"
        if opts.header_fields == 'all':
            # TODO: might vary across file types, thus prior sensing
            # would be needed
            header_fields = h.keys()
        else:
            header_fields = opts.header_fields.split(',')

        for f in header_fields:
            if not f:  # skip empty
                continue
            try:
                row += [str(h[f])]
            except (KeyError, ValueError):
                row += [_err()]

    try:
        if (
            hasattr(h, 'get_qform')
            and hasattr(h, 'get_sform')
            and (h.get_qform() != h.get_sform()).any()
        ):
            row += ['sform']
        else:
            row += ['']
    except Exception as e:
        verbose(2, f'Failed to obtain qform or sform -- {e}')
        if isinstance(h, nib.AnalyzeHeader):
            row += ['']
        else:
            row += [_err()]

    if opts.stats or opts.counts:
        # We are doomed to load data
        try:
            d = np.asarray(vol.dataobj)
            if not opts.stats_zeros:
                d = d[np.nonzero(d)]
            else:
                # at least flatten it -- functionality below doesn't
                # depend on the original shape, so let's use a flat view
                d = d.reshape(-1)
            if opts.stats:
                # just # of elements
                row += [f'@l[{np.prod(d.shape)}]']
                # stats
                row += [f'@l[{np.min(d):.2g}, {np.max(d):.2g}]' if len(d) else '-']
            if opts.counts:
                items, inv = np.unique(d, return_inverse=True)
                if len(items) > 1000 and not opts.all_counts:
                    counts = _err(f'{len(items)} uniques. Use --all-counts')
                else:
                    freq = np.bincount(inv)
                    counts = ' '.join(f'{i:g}:{f}' for i, f in zip(items, freq))
                row += ['@l' + counts]
        except OSError as e:
            verbose(2, f'Failed to obtain stats/counts -- {e}')
            row += [_err()]
    return row


def main(args=None):
    """Show must go on"""

    parser = get_opt_parser()
    (opts, files) = parser.parse_args(args=args)

    nibabel.cmdline.utils.verbose_level = opts.verbose

    if nibabel.cmdline.utils.verbose_level < 3:
        # suppress nibabel format-compliance warnings
        nib.imageglobals.logger.level = 50

    rows = [proc_file(f, opts) for f in files]

    print(table2string(rows))



"""Test we can correctly import example PARREC files"""

import unittest
from glob import glob
from os.path import basename, exists, splitext
from os.path import join as pjoin

import numpy as np
from numpy.testing import assert_almost_equal

from .. import load as top_load
from ..affines import voxel_sizes
from ..parrec import load
from .nibabel_data import get_nibabel_data, needs_nibabel_data

BALLS = pjoin(get_nibabel_data(), 'nitest-balls1')
OBLIQUE = pjoin(get_nibabel_data(), 'parrec_oblique')

# Amount by which affine translation differs from NIFTI conversion
AFF_OFF = [-0.93575081, -0.95657335, 0.03264122]


@needs_nibabel_data('nitest-balls1')
def test_loading():
    # Test loading of parrec files
    for par in glob(pjoin(BALLS, 'PARREC', '*.PAR')):
        par_root, ext = splitext(basename(par))
        # NA.PAR appears to be a localizer, with three slices in each of the
        # three orientations: sagittal; coronal, transverse
        if par_root == 'NA':
            continue
        # Check we can load the image
        pimg = load(par)
        assert pimg.shape[:3] == (80, 80, 10)
        # Compare against NIFTI if present
        nifti_fname = pjoin(BALLS, 'NIFTI', par_root + '.nii.gz')
        if exists(nifti_fname):
            nimg = top_load(nifti_fname)
            assert_almost_equal(nimg.affine[:3, :3], pimg.affine[:3, :3], 3)
            # The translation part is always off by the same ammout
            aff_off = pimg.affine[:3, 3] - nimg.affine[:3, 3]
            assert_almost_equal(aff_off, AFF_OFF, 4)
            # The difference is max in the order of 0.5 voxel
            vox_sizes = voxel_sizes(nimg.affine)
            assert np.all(np.abs(aff_off / vox_sizes) <= 0.501)
            # The data is very close, unless it's the fieldmap
            if par_root != 'fieldmap':
                assert np.allclose(pimg.dataobj, nimg.dataobj)
            # Not sure what's going on with the fieldmap image - TBA


@needs_nibabel_data('nitest-balls1')
def test_fieldmap():
    # Test fieldmap image
    # Exploring the DICOM suggests that the first volume is magnitude and the
    # second is phase.  The NIfTI has very odd scaling, being all negative.
    fieldmap_par = pjoin(BALLS, 'PARREC', 'fieldmap.PAR')
    fieldmap_nii = pjoin(BALLS, 'NIFTI', 'fieldmap.nii.gz')
    load(fieldmap_par)
    top_load(fieldmap_nii)
    raise unittest.SkipTest('Fieldmap remains puzzling')


@needs_nibabel_data('parrec_oblique')
def test_oblique_loading():
    # Test loading of oblique parrec files
    for par in glob(pjoin(OBLIQUE, 'PARREC', '*.PAR')):
        par_root, ext = splitext(basename(par))
        # Check we can load the image
        pimg = load(par)
        assert pimg.shape == (560, 560, 1)
        # Compare against NIFTI if present
        nifti_fname = pjoin(OBLIQUE, 'NIFTI', par_root + '.nii')
        nimg = top_load(nifti_fname)
        assert_almost_equal(nimg.affine[:3, :3], pimg.affine[:3, :3], 3)
        # The translation part is always off
        # The amount differs by rotation
        aff_off = pimg.affine[:3, 3] - nimg.affine[:3, 3]
        # The difference is max in the order of 0.5 voxel
        vox_sizes = voxel_sizes(nimg.affine)
        assert np.all(np.abs(aff_off / vox_sizes) <= 0.5)
        # The data is very close
        assert np.allclose(pimg.dataobj, nimg.dataobj)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Test diff"""

from os.path import abspath, dirname
from os.path import join as pjoin

import numpy as np

DATA_PATH = abspath(pjoin(dirname(__file__), 'data'))

from nibabel.cmdline.diff import are_values_different


def test_diff_values_int():
    large = 10**30
    assert not are_values_different(0, 0)
    assert not are_values_different(1, 1)
    assert not are_values_different(large, large)
    assert are_values_different(0, 1)
    assert are_values_different(1, 2)
    assert are_values_different(1, large)


def test_diff_values_float():
    assert not are_values_different(0.0, 0.0)
    assert not are_values_different(0.0, 0.0, 0.0)  # can take more
    assert not are_values_different(1.1, 1.1)
    assert are_values_different(0.0, 1.1)
    assert are_values_different(0.0, 0, 1.1)
    assert are_values_different(1.0, 2.0)


def test_diff_values_mixed():
    assert are_values_different(1.0, 1)
    assert are_values_different(1.0, '1')
    assert are_values_different(1, '1')
    assert are_values_different(1, None)
    assert are_values_different(np.ndarray([0]), 'hey')
    assert not are_values_different(None, None)


def test_diff_values_array():
    from numpy import array, inf, nan

    a_int = array([1, 2])
    a_float = a_int.astype(float)

    assert are_values_different(a_int, a_float)
    assert are_values_different(a_int, a_int, a_float)
    assert are_values_different(np.arange(3), np.arange(1, 4))
    assert are_values_different(np.arange(3), np.arange(4))
    assert are_values_different(np.arange(4), np.arange(4).reshape((2, 2)))
    # no broadcasting should kick in - shape difference
    assert are_values_different(array([1]), array([1, 1]))
    assert not are_values_different(a_int, a_int)
    assert not are_values_different(a_float, a_float)

    # nans - we consider them "the same" for the purpose of these comparisons
    assert not are_values_different(nan, nan)
    assert not are_values_different(nan, nan, nan)
    assert are_values_different(nan, nan, 1)
    assert are_values_different(1, nan, nan)
    assert not are_values_different(array([nan, nan]), array([nan, nan]))
    assert not are_values_different(array([nan, nan]), array([nan, nan]), array([nan, nan]))
    assert not are_values_different(array([nan, 1]), array([nan, 1]))
    assert are_values_different(array([nan, nan]), array([nan, 1]))
    assert are_values_different(array([0, nan]), array([nan, 0]))
    assert are_values_different(array([1, 2, 3, nan]), array([nan, 3, 5, 4]))
    assert are_values_different(nan, 1.0)
    assert are_values_different(array([1, 2, 3, nan]), array([3, 4, 5, nan]))
    # and some inf should not be a problem
    assert not are_values_different(array([0, inf]), array([0, inf]))
    assert are_values_different(array([0, inf]), array([inf, 0]))

    # we will allow for types to be of different endianness but the
    # same in "instantiation" type and value
    assert not are_values_different(np.array(1, dtype='<i4'), np.array(1, dtype='>i4'))
    # but do report difference if instantiation type is different:
    assert are_values_different(np.array(1, dtype='<i4'), np.array(1, dtype='<i2'))



"""Just showing the mosaic simplification"""

from sympy import Matrix, Symbol, simplify, symbols


def numbered_matrix(nrows, ncols, symbol_prefix):
    return Matrix(nrows, ncols, lambda i, j: Symbol(symbol_prefix + '_{%d%d}' % (i + 1, j + 1)))


def numbered_vector(nrows, symbol_prefix):
    return Matrix(nrows, 1, lambda i, j: Symbol(symbol_prefix + '_{%d}' % (i + 1)))


RS = numbered_matrix(3, 3, 'rs')

mdc, mdr, rdc, rdr = symbols('md_{cols} md_{rows} rd_{cols} rd_{rows}')

md_adj = Matrix((mdc - 1, mdr - 1, 0)) / -2
rd_adj = Matrix((rdc - 1, rdr - 1, 0)) / -2

adj = -(RS * md_adj) + RS * rd_adj
adj.simplify()

Q = RS[:, :2] * Matrix(((mdc - rdc) / 2, (mdr - rdr) / 2))

assert simplify(adj - Q) == Matrix([0, 0, 0])



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Utility routines for working with points and affine transforms"""

from functools import reduce

import numpy as np


class AffineError(ValueError):
    """Errors in calculating or using affines"""

    # Inherits from ValueError to keep compatibility with ValueError previously
    # raised in append_diag
    pass


def apply_affine(aff, pts, inplace=False):
    """Apply affine matrix `aff` to points `pts`

    Returns result of application of `aff` to the *right* of `pts`.  The
    coordinate dimension of `pts` should be the last.

    For the 3D case, `aff` will be shape (4,4) and `pts` will have final axis
    length 3 - maybe it will just be N by 3. The return value is the
    transformed points, in this case::

        res = np.dot(aff[:3,:3], pts.T) + aff[:3,3:4]
        transformed_pts = res.T

    This routine is more general than 3D, in that `aff` can have any shape
    (N,N), and `pts` can have any shape, as long as the last dimension is for
    the coordinates, and is therefore length N-1.

    Parameters
    ----------
    aff : (N, N) array-like
        Homogeneous affine, for 3D points, will be 4 by 4. Contrary to first
        appearance, the affine will be applied on the left of `pts`.
    pts : (..., N-1) array-like
        Points, where the last dimension contains the coordinates of each
        point.  For 3D, the last dimension will be length 3.
    inplace : bool, optional
        If True, attempt to apply the affine directly to ``pts``.
        If False, or in-place application fails, a freshly allocated
        array will be returned.

    Returns
    -------
    transformed_pts : (..., N-1) array
        transformed points

    Examples
    --------
    >>> aff = np.array([[0,2,0,10],[3,0,0,11],[0,0,4,12],[0,0,0,1]])
    >>> pts = np.array([[1,2,3],[2,3,4],[4,5,6],[6,7,8]])
    >>> apply_affine(aff, pts) #doctest: +ELLIPSIS
    array([[14, 14, 24],
           [16, 17, 28],
           [20, 23, 36],
           [24, 29, 44]]...)

    Just to show that in the simple 3D case, it is equivalent to:

    >>> (np.dot(aff[:3,:3], pts.T) + aff[:3,3:4]).T #doctest: +ELLIPSIS
    array([[14, 14, 24],
           [16, 17, 28],
           [20, 23, 36],
           [24, 29, 44]]...)

    But `pts` can be a more complicated shape:

    >>> pts = pts.reshape((2,2,3))
    >>> apply_affine(aff, pts) #doctest: +ELLIPSIS
    array([[[14, 14, 24],
            [16, 17, 28]],
    <BLANKLINE>
           [[20, 23, 36],
            [24, 29, 44]]]...)
    """
    aff = np.asarray(aff)
    pts = np.asarray(pts)
    shape = pts.shape
    pts = pts.reshape((-1, shape[-1]))
    # rzs == rotations, zooms, shears
    rzs = aff[:-1, :-1]
    trans = aff[:-1, -1]

    if inplace:
        try:
            np.dot(pts, rzs.T, out=pts)
        except ValueError:
            inplace = False
        else:
            pts += trans[None, :]
    if not inplace:
        pts = pts @ rzs.T + trans[None, :]

    return pts.reshape(shape)


def to_matvec(transform):
    """Split a transform into its matrix and vector components

    The transformation must be represented in homogeneous coordinates and is
    split into its rotation matrix and translation vector components.

    Parameters
    ----------
    transform : array-like
        NxM transform matrix in homogeneous coordinates representing an affine
        transformation from an (N-1)-dimensional space to an (M-1)-dimensional
        space. An example is a 4x4 transform representing rotations and
        translations in 3 dimensions. A 4x3 matrix can represent a
        2-dimensional plane embedded in 3 dimensional space.

    Returns
    -------
    matrix : (N-1, M-1) array
        Matrix component of `transform`
    vector : (M-1,) array
        Vector component of `transform`

    See Also
    --------
    from_matvec

    Examples
    --------
    >>> aff = np.diag([2, 3, 4, 1])
    >>> aff[:3,3] = [9, 10, 11]
    >>> to_matvec(aff)
    (array([[2, 0, 0],
           [0, 3, 0],
           [0, 0, 4]]), array([ 9, 10, 11]))
    """
    transform = np.asarray(transform)
    ndimin = transform.shape[0] - 1
    ndimout = transform.shape[1] - 1
    matrix = transform[0:ndimin, 0:ndimout]
    vector = transform[0:ndimin, ndimout]
    return matrix, vector


def from_matvec(matrix, vector=None):
    """Combine a matrix and vector into an homogeneous affine

    Combine a rotation / scaling / shearing matrix and translation vector into
    a transform in homogeneous coordinates.

    Parameters
    ----------
    matrix : array-like
        An NxM array representing the the linear part of the transform.
        A transform from an M-dimensional space to an N-dimensional space.
    vector : None or array-like, optional
        None or an (N,) array representing the translation. None corresponds to
        an (N,) array of zeros.

    Returns
    -------
    xform : array
        An (N+1, M+1) homogeneous transform matrix.

    See Also
    --------
    to_matvec

    Examples
    --------
    >>> from_matvec(np.diag([2, 3, 4]), [9, 10, 11])
    array([[ 2,  0,  0,  9],
           [ 0,  3,  0, 10],
           [ 0,  0,  4, 11],
           [ 0,  0,  0,  1]])

    The `vector` argument is optional:

    >>> from_matvec(np.diag([2, 3, 4]))
    array([[2, 0, 0, 0],
           [0, 3, 0, 0],
           [0, 0, 4, 0],
           [0, 0, 0, 1]])
    """
    matrix = np.asarray(matrix)
    nin, nout = matrix.shape
    t = np.zeros((nin + 1, nout + 1), matrix.dtype)
    t[0:nin, 0:nout] = matrix
    t[nin, nout] = 1.0
    if vector is not None:
        t[0:nin, nout] = vector
    return t


def append_diag(aff, steps, starts=()):
    """Add diagonal elements `steps` and translations `starts` to affine

    Typical use is in expanding 4x4 affines to larger dimensions.  Nipy is the
    main consumer because it uses NxM affines, whereas we generally only use
    4x4 affines; the routine is here for convenience.

    Parameters
    ----------
    aff : 2D array
        N by M affine matrix
    steps : scalar or sequence
        diagonal elements to append.
    starts : scalar or sequence
        elements to append to last column of `aff`, representing translations
        corresponding to the `steps`. If empty, expands to a vector of zeros
        of the same length as `steps`

    Returns
    -------
    aff_plus : 2D array
        Now P by Q where L = ``len(steps)`` and P == N+L, Q=N+L

    Examples
    --------
    >>> aff = np.eye(4)
    >>> aff[:3,:3] = np.arange(9).reshape((3,3))
    >>> append_diag(aff, [9, 10], [99,100])
    array([[  0.,   1.,   2.,   0.,   0.,   0.],
           [  3.,   4.,   5.,   0.,   0.,   0.],
           [  6.,   7.,   8.,   0.,   0.,   0.],
           [  0.,   0.,   0.,   9.,   0.,  99.],
           [  0.,   0.,   0.,   0.,  10., 100.],
           [  0.,   0.,   0.,   0.,   0.,   1.]])
    """
    aff = np.asarray(aff)
    steps = np.atleast_1d(steps)
    starts = np.atleast_1d(starts)
    n_steps = len(steps)
    if len(starts) == 0:
        starts = np.zeros(n_steps, dtype=steps.dtype)
    elif len(starts) != n_steps:
        raise AffineError('Steps should have same length as starts')
    old_n_out, old_n_in = aff.shape[0] - 1, aff.shape[1] - 1
    # make new affine
    aff_plus = np.zeros((old_n_out + n_steps + 1, old_n_in + n_steps + 1), dtype=aff.dtype)
    # Get stuff from old affine
    aff_plus[:old_n_out, :old_n_in] = aff[:old_n_out, :old_n_in]
    aff_plus[:old_n_out, -1] = aff[:old_n_out, -1]
    # Add new diagonal elements
    for i, el in enumerate(steps):
        aff_plus[old_n_out + i, old_n_in + i] = el
    # Add translations for new affine, plus last 1
    aff_plus[old_n_out:, -1] = list(starts) + [1]
    return aff_plus


def dot_reduce(*args):
    r"""Apply numpy dot product function from right to left on arrays

    For passed arrays :math:`A, B, C, ... Z` returns :math:`A \dot B \dot C ...
    \dot Z` where "." is the numpy array dot product.

    Parameters
    ----------
    \*\*args : arrays
        Arrays that can be passed to numpy ``dot`` function

    Returns
    -------
    dot_product : array
        If there are N arguments, result of ``arg[0].dot(arg[1].dot(arg[2].dot
        ...  arg[N-2].dot(arg[N-1])))...``
    """
    return reduce(lambda x, y: np.dot(y, x), args[::-1])


def voxel_sizes(affine):
    r"""Return voxel size for each input axis given `affine`

    The `affine` is the mapping between array (voxel) coordinates and mm
    (world) coordinates.

    The voxel size for the first voxel (array) axis is the distance moved in
    world coordinates when moving one unit along the first voxel (array) axis.
    This is the distance between the world coordinate of voxel (0, 0, 0) and
    the world coordinate of voxel (1, 0, 0).  The world coordinate vector of
    voxel coordinate vector (0, 0, 0) is given by ``v0 = affine.dot((0, 0, 0,
    1)[:3]``.  The world coordinate vector of voxel vector (1, 0, 0) is
    ``v1_ax1 = affine.dot((1, 0, 0, 1))[:3]``.  The final 1 in the voxel
    vectors and the ``[:3]`` at the end are because the affine works on
    homogeneous coordinates.  The translations part of the affine is ``trans =
    affine[:3, 3]``, and the rotations, zooms and shearing part of the affine
    is ``rzs = affine[:3, :3]``. Because of the final 1 in the input voxel
    vector, ``v0 == rzs.dot((0, 0, 0)) + trans``, and ``v1_ax1 == rzs.dot((1,
    0, 0)) + trans``, and the difference vector is ``rzs.dot((0, 0, 0)) -
    rzs.dot((1, 0, 0)) == rzs.dot((1, 0, 0)) == rzs[:, 0]``.  The distance
    vectors in world coordinates between (0, 0, 0) and (1, 0, 0), (0, 1, 0),
    (0, 0, 1) are given by ``rzs.dot(np.eye(3)) = rzs``.  The voxel sizes are
    the Euclidean lengths of the distance vectors.  So, the voxel sizes are
    the Euclidean lengths of the columns of the affine (excluding the last row
    and column of the affine).

    Parameters
    ----------
    affine : 2D array-like
        Affine transformation array.  Usually shape (4, 4), but can be any 2D
        array.

    Returns
    -------
    vox_sizes : 1D array
        Voxel sizes for each input axis of affine.  Usually 1D array length 3,
        but in general has length (N-1) where input `affine` is shape (M, N).
    """
    top_left = affine[:-1, :-1]
    return np.sqrt(np.sum(top_left**2, axis=0))


def obliquity(affine):
    r"""Estimate the *obliquity* an affine's axes represent

    The term *obliquity* is defined here as the rotation of those axes with
    respect to the cardinal axes.
    This implementation is inspired by `AFNI's implementation
    <https://github.com/afni/afni/blob/b6a9f7a21c1f3231ff09efbd861f8975ad48e525/src/thd_coords.c#L660-L698>`_.
    For further details about *obliquity*, check `AFNI's documentation
    <https://sscc.nimh.nih.gov/sscc/dglen/Obliquity>`_.

    Parameters
    ----------
    affine : 2D array-like
        Affine transformation array.  Usually shape (4, 4), but can be any 2D
        array.

    Returns
    -------
    angles : 1D array-like
        The *obliquity* of each axis with respect to the cardinal axes, in radians.

    """
    vs = voxel_sizes(affine)
    best_cosines = np.abs(affine[:-1, :-1] / vs).max(axis=1)
    return np.arccos(best_cosines)


def rescale_affine(affine, shape, zooms, new_shape=None):
    """Return a new affine matrix with updated voxel sizes (zooms)

    This function preserves the rotations and shears of the original
    affine, as well as the RAS location of the central voxel of the
    image.

    Parameters
    ----------
    affine : (N, N) array-like
        NxN transform matrix in homogeneous coordinates representing an affine
        transformation from an (N-1)-dimensional space to an (N-1)-dimensional
        space. An example is a 4x4 transform representing rotations and
        translations in 3 dimensions.
    shape : (N-1,) array-like
        The extent of the (N-1) dimensions of the original space
    zooms : (N-1,) array-like
        The size of voxels of the output affine
    new_shape : (N-1,) array-like, optional
        The extent of the (N-1) dimensions of the space described by the
        new affine. If ``None``, use ``shape``.

    Returns
    -------
    affine : (N, N) array
        A new affine transform with the specified voxel sizes

    """
    shape = np.asarray(shape)
    new_shape = np.array(new_shape if new_shape is not None else shape)

    s = voxel_sizes(affine)
    rzs_out = affine[:3, :3] * zooms / s

    # Using xyz = A @ ijk, determine translation
    centroid = apply_affine(affine, (shape - 1) // 2)
    t_out = centroid - rzs_out @ ((new_shape - 1) // 2)
    return from_matvec(rzs_out, t_out)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Array proxy base class

The proxy API is - at minimum:

* The object has a read-only attribute ``shape``
* read only ``is_proxy`` attribute / property set to True
* the object returns the data array from ``np.asarray(prox)``
* returns array slice from ``prox[<slice_spec>]`` where ``<slice_spec>`` is any
  ndarray slice specification that does not use numpy 'advanced indexing'.
* modifying no object outside ``obj`` will affect the result of
  ``np.asarray(obj)``.  Specifically:

  * Changes in position (``obj.tell()``) of passed file-like objects will
    not affect the output of from ``np.asarray(proxy)``.
  * if you pass a header into the __init__, then modifying the original
    header will not affect the result of the array return.

See :mod:`nibabel.tests.test_proxy_api` for proxy API conformance checks.
"""

from __future__ import annotations

import typing as ty
import warnings
from contextlib import contextmanager
from threading import RLock

import numpy as np

from . import _compression, openers
from .fileslice import canonical_slicers, fileslice
from .volumeutils import apply_read_scaling, array_from_file

"""This flag controls whether a new file handle is created every time an image
is accessed through an ``ArrayProxy``, or a single file handle is created and
used for the lifetime of the ``ArrayProxy``. It should be set to one of
``True`` or ``False``.

Management of file handles will be performed either by ``ArrayProxy`` objects,
or by the ``indexed_gzip`` package if it is used.

If this flag is set to ``True``, a single file handle is created and used. If
``False``, a new file handle is created every time the image is accessed.

If this is set to any other value, attempts to create an ``ArrayProxy`` without
specifying the ``keep_file_open`` flag will result in a ``ValueError`` being
raised.
"""
KEEP_FILE_OPEN_DEFAULT = False


if ty.TYPE_CHECKING:
    import numpy.typing as npt

    from ._typing import Self, TypeVar

    # Taken from numpy/__init__.pyi
    _DType = TypeVar('_DType', bound=np.dtype[ty.Any])


class ArrayLike(ty.Protocol):
    """Protocol for numpy ndarray-like objects

    This is more stringent than :class:`numpy.typing.ArrayLike`, but guarantees
    access to shape, ndim and slicing.
    """

    shape: tuple[int, ...]

    @property
    def ndim(self) -> int: ...

    # If no dtype is passed, any dtype might be returned, depending on the array-like
    @ty.overload
    def __array__(self, dtype: None = ..., /) -> np.ndarray[ty.Any, np.dtype[ty.Any]]: ...

    # Any dtype might be passed, and *that* dtype must be returned
    @ty.overload
    def __array__(self, dtype: _DType, /) -> np.ndarray[ty.Any, _DType]: ...

    def __getitem__(self, key, /) -> npt.NDArray: ...


class ArrayProxy(ArrayLike):
    """Class to act as proxy for the array that can be read from a file

    The array proxy allows us to freeze the passed fileobj and header such that
    it returns the expected data array.

    This implementation assumes a contiguous array in the file object, with one
    of the numpy dtypes, starting at a given file position ``offset`` with
    single ``slope`` and ``intercept`` scaling to produce output values.

    The class ``__init__`` requires a spec which defines how the data will be
    read and rescaled. The spec may be a tuple of length 2 - 5, containing the
    shape, storage dtype, offset, slope and intercept, or a ``header`` object
    with methods:

    * get_data_shape
    * get_data_dtype
    * get_data_offset
    * get_slope_inter

    A header should also have a 'copy' method.  This requirement will go away
    when the deprecated 'header' property goes away.

    This implementation allows us to deal with Analyze and its variants,
    including Nifti1, and with the MGH format.

    Other image types might need more specific classes to implement the API.
    See :mod:`nibabel.minc1`, :mod:`nibabel.ecat` and :mod:`nibabel.parrec` for
    examples.
    """

    _default_order = 'F'

    def __init__(self, file_like, spec, *, mmap=True, order=None, keep_file_open=None):
        """Initialize array proxy instance

        Parameters
        ----------
        file_like : object
            File-like object or filename. If file-like object, should implement
            at least ``read`` and ``seek``.
        spec : object or tuple
            Tuple must have length 2-5, with the following values:

            #. shape: tuple - tuple of ints describing shape of data;
            #. storage_dtype: dtype specifier - dtype of array inside proxied
               file, or input to ``numpy.dtype`` to specify array dtype;
            #. offset: int - offset, in bytes, of data array from start of file
               (default: 0);
            #. slope: float - scaling factor for resulting data (default: 1.0);
            #. inter: float - intercept for rescaled data (default: 0.0).

            OR

            Header object implementing ``get_data_shape``, ``get_data_dtype``,
            ``get_data_offset``, ``get_slope_inter``
        mmap : {True, False, 'c', 'r'}, optional, keyword only
            `mmap` controls the use of numpy memory mapping for reading data.
            If False, do not try numpy ``memmap`` for data array.  If one of
            {'c', 'r'}, try numpy memmap with ``mode=mmap``.  A `mmap` value of
            True gives the same behavior as ``mmap='c'``.  If `file_like`
            cannot be memory-mapped, ignore `mmap` value and read array from
            file.
        order : {None, 'F', 'C'}, optional, keyword only
            `order` controls the order of the data array layout. Fortran-style,
            column-major order may be indicated with 'F', and C-style, row-major
            order may be indicated with 'C'. None gives the default order, that
            comes from the `_default_order` class variable.
        keep_file_open : { None, True, False }, optional, keyword only
            `keep_file_open` controls whether a new file handle is created
            every time the image is accessed, or a single file handle is
            created and used for the lifetime of this ``ArrayProxy``. If
            ``True``, a single file handle is created and used. If ``False``,
            a new file handle is created every time the image is accessed.
            If ``file_like`` is an open file handle, this setting has no
            effect. The default value (``None``) will result in the value of
            ``KEEP_FILE_OPEN_DEFAULT`` being used.
        """
        if mmap not in (True, False, 'c', 'r'):
            raise ValueError("mmap should be one of {True, False, 'c', 'r'}")
        if order not in (None, 'C', 'F'):
            raise ValueError("order should be one of {None, 'C', 'F'}")
        self.file_like = file_like
        if hasattr(spec, 'get_data_shape'):
            slope, inter = spec.get_slope_inter()
            par = (
                spec.get_data_shape(),
                spec.get_data_dtype(),
                spec.get_data_offset(),
                1.0 if slope is None else slope,
                0.0 if inter is None else inter,
            )
        elif 2 <= len(spec) <= 5:
            optional = (0, 1.0, 0.0)
            par = spec + optional[len(spec) - 2 :]
        else:
            raise TypeError('spec must be tuple of length 2-5 or header object')

        # Warn downstream users that the class variable order is going away
        if hasattr(self.__class__, 'order'):
            warnings.warn(
                f'Class {self.__class__} has an `order` class variable. '
                'ArrayProxy subclasses should rename this variable to `_default_order` '
                'to avoid conflict with instance variables.\n'
                '* deprecated in version: 5.0\n'
                '* will raise error in version: 7.0\n',
                DeprecationWarning,
                stacklevel=2,
            )
            # Override _default_order with order, to follow intent of subclasser
            self._default_order = self.order

        # Copies of values needed to read array
        self._shape, self._dtype, self._offset, self._slope, self._inter = par
        # Permit any specifier that can be interpreted as a numpy dtype
        self._dtype = np.dtype(self._dtype)
        self._mmap = mmap
        if order is None:
            order = self._default_order
        self.order = order
        # Flags to keep track of whether a single ImageOpener is created, and
        # whether a single underlying file handle is created.
        self._keep_file_open, self._persist_opener = self._should_keep_file_open(keep_file_open)
        self._lock = RLock()

    def _has_fh(self) -> bool:
        """Determine if our file-like is a filehandle or path"""
        return hasattr(self.file_like, 'read') and hasattr(self.file_like, 'seek')

    def copy(self) -> Self:
        """Create a new ArrayProxy for the same file and parameters

        If the proxied file is an open file handle, the new ArrayProxy
        will share a lock with the old one.
        """
        spec = self._shape, self._dtype, self._offset, self._slope, self._inter
        new = self.__class__(
            self.file_like,
            spec,
            mmap=self._mmap,
            keep_file_open=self._keep_file_open,
        )
        if self._has_fh():
            new._lock = self._lock
        return new

    def __del__(self):
        """If this ``ArrayProxy`` was created with ``keep_file_open=True``,
        the open file object is closed if necessary.
        """
        if hasattr(self, '_opener') and not self._opener.closed:
            self._opener.close_if_mine()
            self._opener = None

    def __getstate__(self):
        """Returns the state of this ``ArrayProxy`` during pickling."""
        state = self.__dict__.copy()
        state.pop('_lock', None)
        return state

    def __setstate__(self, state):
        """Sets the state of this ``ArrayProxy`` during unpickling."""
        self.__dict__.update(state)
        self._lock = RLock()

    def _should_keep_file_open(self, keep_file_open):
        """Called by ``__init__``.

        This method determines how to manage ``ImageOpener`` instances,
        and the underlying file handles - the behaviour depends on:

         - whether ``self.file_like`` is an an open file handle, or a path to a
           ``'.gz'`` file, or a path to a non-gzip file.
         - whether ``indexed_gzip`` is present (see
           :attr:`._compression.HAVE_INDEXED_GZIP`).

        An ``ArrayProxy`` object uses two internal flags to manage
        ``ImageOpener`` instances and underlying file handles.

          - The ``_persist_opener`` flag controls whether a single
            ``ImageOpener`` should be created and used for the lifetime of
            this ``ArrayProxy``, or whether separate ``ImageOpener`` instances
            should be created on each file access.

          - The ``_keep_file_open`` flag controls qwhether the underlying file
            handle should be kept open for the lifetime of this
            ``ArrayProxy``, or whether the file handle should be (re-)opened
            and closed on each file access.

        The internal ``_keep_file_open`` flag is only relevant if
        ``self.file_like`` is a ``'.gz'`` file, and the ``indexed_gzip`` library is
        present.

        This method returns the values to be used for the internal
        ``_persist_opener`` and ``_keep_file_open`` flags; these values are
        derived according to the following rules:

        1. If ``self.file_like`` is a file(-like) object, both flags are set to
        ``False``.

        2. If ``keep_file_open`` (as passed to :meth:``__init__``) is
           ``True``, both internal flags are set to ``True``.

        3. If ``keep_file_open`` is ``False``, but ``self.file_like`` is not a path
           to a ``.gz`` file or ``indexed_gzip`` is not present, both flags
           are set to ``False``.

        4. If ``keep_file_open`` is ``False``, ``self.file_like`` is a path to a
           ``.gz`` file, and ``indexed_gzip`` is present, ``_persist_opener``
           is set to ``True``, and ``_keep_file_open`` is set to ``False``.
           In this case, file handle management is delegated to the
           ``indexed_gzip`` library.

        Parameters
        ----------

        keep_file_open : { True, False }
            Flag as passed to ``__init__``.

        Returns
        -------

        A tuple containing:
          - ``keep_file_open`` flag to control persistence of file handles
          - ``persist_opener`` flag to control persistence of ``ImageOpener``
            objects.
        """
        if keep_file_open is None:
            keep_file_open = KEEP_FILE_OPEN_DEFAULT
            if keep_file_open not in (True, False):
                raise ValueError(
                    'nibabel.arrayproxy.KEEP_FILE_OPEN_DEFAULT '
                    f'must be boolean. Found: {keep_file_open}'
                )
        elif keep_file_open not in (True, False):
            raise ValueError('keep_file_open must be one of {None, True, False}')

        # file_like is a handle - keep_file_open is irrelevant
        if self._has_fh():
            return False, False
        # if the file is a gzip file, and we have_indexed_gzip,
        have_igzip = _compression.HAVE_INDEXED_GZIP and self.file_like.endswith('.gz')

        persist_opener = keep_file_open or have_igzip
        return keep_file_open, persist_opener

    @property
    def shape(self):
        return self._shape

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def dtype(self):
        return self._dtype

    @property
    def offset(self):
        return self._offset

    @property
    def slope(self):
        return self._slope

    @property
    def inter(self):
        return self._inter

    @property
    def is_proxy(self):
        return True

    @contextmanager
    def _get_fileobj(self):
        """Create and return a new ``ImageOpener``, or return an existing one.

        The specific behaviour depends on the value of the ``keep_file_open``
        flag that was passed to ``__init__``.

        Yields
        ------
        ImageOpener
            A newly created ``ImageOpener`` instance, or an existing one,
            which provides access to the file.
        """
        if self._persist_opener:
            if not hasattr(self, '_opener'):
                self._opener = openers.ImageOpener(self.file_like, keep_open=self._keep_file_open)
            yield self._opener
        else:
            with openers.ImageOpener(self.file_like, keep_open=False) as opener:
                yield opener

    def _get_unscaled(self, slicer):
        if canonical_slicers(slicer, self._shape, False) == canonical_slicers(
            (), self._shape, False
        ):
            with self._get_fileobj() as fileobj, self._lock:
                return array_from_file(
                    self._shape,
                    self._dtype,
                    fileobj,
                    offset=self._offset,
                    order=self.order,
                    mmap=self._mmap,
                )
        with self._get_fileobj() as fileobj:
            return fileslice(
                fileobj,
                slicer,
                self._shape,
                self._dtype,
                self._offset,
                order=self.order,
                lock=self._lock,
            )

    def _get_scaled(self, dtype, slicer):
        # Ensure scale factors have dtypes
        scl_slope = np.asanyarray(self._slope)
        scl_inter = np.asanyarray(self._inter)
        use_dtype = scl_slope.dtype if dtype is None else dtype

        if np.can_cast(scl_slope, use_dtype):
            scl_slope = scl_slope.astype(use_dtype)
        if np.can_cast(scl_inter, use_dtype):
            scl_inter = scl_inter.astype(use_dtype)
        # Read array and upcast as necessary for big slopes, intercepts
        scaled = apply_read_scaling(self._get_unscaled(slicer=slicer), scl_slope, scl_inter)
        if dtype is not None:
            scaled = scaled.astype(np.promote_types(scaled.dtype, dtype), copy=False)
        return scaled

    def get_unscaled(self):
        """Read data from file

        This is an optional part of the proxy API
        """
        return self._get_unscaled(slicer=())

    def __array__(self, dtype=None):
        """Read data from file and apply scaling, casting to ``dtype``

        If ``dtype`` is unspecified, the dtype of the returned array is the
        narrowest dtype that can represent the data without overflow.
        Generally, it is the wider of the dtypes of the slopes or intercepts.

        The types of the scale factors will generally be determined by the
        parameter size in the image header, and so should be consistent for a
        given image format, but may vary across formats.

        Parameters
        ----------
        dtype : numpy dtype specifier, optional
            A numpy dtype specifier specifying the type of the returned array.

        Returns
        -------
        array
            Scaled image data with type `dtype`.
        """
        arr = self._get_scaled(dtype=dtype, slicer=())
        if dtype is not None:
            arr = arr.astype(dtype, copy=False)
        return arr

    def __getitem__(self, slicer):
        return self._get_scaled(dtype=None, slicer=slicer)

    def reshape(self, shape):
        """Return an ArrayProxy with a new shape, without modifying data"""
        size = np.prod(self._shape)

        # Calculate new shape if not fully specified
        from functools import reduce
        from operator import mul

        n_unknowns = len([e for e in shape if e == -1])
        if n_unknowns > 1:
            raise ValueError('can only specify one unknown dimension')
        elif n_unknowns == 1:
            known_size = reduce(mul, shape, -1)
            unknown_size = size // known_size
            shape = tuple(unknown_size if e == -1 else e for e in shape)

        if np.prod(shape) != size:
            raise ValueError(f'cannot reshape array of size {size:d} into shape {shape!s}')
        return self.__class__(
            file_like=self.file_like,
            spec=(shape, self._dtype, self._offset, self._slope, self._inter),
            mmap=self._mmap,
        )


def is_proxy(obj):
    """Return True if `obj` is an array proxy"""
    try:
        return obj.is_proxy
    except AttributeError:
        return False


def reshape_dataobj(obj, shape):
    """Use `obj` reshape method if possible, else numpy reshape function"""
    return obj.reshape(shape) if hasattr(obj, 'reshape') else np.reshape(obj, shape)


def get_obj_dtype(obj):
    """Get the effective dtype of an array-like object"""
    if is_proxy(obj):
        # Read and potentially apply scaling to one value
        idx = (0,) * len(obj.shape)
        return obj[idx].dtype
    elif hasattr(obj, 'dtype'):
        # Trust the dtype (probably an ndarray)
        return obj.dtype
    else:
        # Coerce; this could be expensive but we don't know what we can do with it
        return np.asanyarray(obj).dtype



import pathlib
import unittest
from importlib.resources import files
from unittest import mock

import pytest

import nibabel as nib


@pytest.mark.parametrize(
    ('verbose', 'v_args'), [(-2, ['-qq']), (-1, ['-q']), (0, []), (1, ['-v']), (2, ['-vv'])]
)
@pytest.mark.parametrize('doctests', (True, False))
@pytest.mark.parametrize('coverage', (True, False))
def test_nibabel_test(verbose, v_args, doctests, coverage):
    expected_args = v_args + ['--doctest-modules', '--cov', 'nibabel', '--pyargs', 'nibabel']
    if not doctests:
        expected_args.remove('--doctest-modules')
    if not coverage:
        expected_args[-4:-2] = []

    with mock.patch('pytest.main') as pytest_main:
        nib.test(verbose=verbose, doctests=doctests, coverage=coverage)

    args, kwargs = pytest_main.call_args
    assert args == ()
    assert kwargs == {'args': expected_args}


def test_nibabel_test_errors():
    with pytest.raises(NotImplementedError):
        nib.test(label='fast')
    with pytest.raises(NotImplementedError):
        nib.test(raise_warnings=[])
    with pytest.raises(NotImplementedError):
        nib.test(timer=True)
    with pytest.raises(ValueError):
        nib.test(verbose='-v')


def test_nibabel_bench():
    config_path = files('nibabel') / 'benchmarks/pytest.benchmark.ini'
    if not isinstance(config_path, pathlib.Path):
        raise unittest.SkipTest('Package is not unpacked; could get temp path')

    expected_args = ['-c', str(config_path), '--pyargs', 'nibabel']

    with mock.patch('pytest.main') as pytest_main:
        nib.bench(verbose=0)

    args, kwargs = pytest_main.call_args
    assert args == ()
    assert kwargs == {'args': expected_args}

    with mock.patch('pytest.main') as pytest_main:
        nib.bench(verbose=0, extra_argv=[])

    args, kwargs = pytest_main.call_args
    assert args == ()
    assert kwargs == {'args': expected_args}



import itertools

import nibabel


def get_affine_from_reference(ref):
    """Returns the affine defining the reference space.

    Parameters
    ----------
    ref : str or :class:`Nifti1Image` object or ndarray shape (4, 4)
        If str then it's the filename of reference file that will be loaded
        using :func:`nibabel.load` in order to obtain the affine.
        If :class:`Nifti1Image` object then the affine is obtained from it.
        If ndarray shape (4, 4) then it's the affine.

    Returns
    -------
    affine : ndarray (4, 4)
        Transformation matrix mapping voxel space to RAS+mm space.
    """
    if hasattr(ref, 'affine'):
        return ref.affine

    if hasattr(ref, 'shape'):
        if ref.shape != (4, 4):
            msg = '`ref` needs to be a numpy array with shape (4, 4)!'
            raise ValueError(msg)

        return ref

    # Assume `ref` is the name of a neuroimaging file.
    return nibabel.load(ref).affine


def peek_next(iterable):
    """Peek next element of iterable.

    Parameters
    ----------
    iterable
        Iterable to peek the next element from.

    Returns
    -------
    next_item
        Element peeked from `iterable`.
    new_iterable
        Iterable behaving like if the original `iterable` was untouched.
    """
    next_item = next(iterable)
    return next_item, itertools.chain([next_item], iterable)



#!/usr/bin/env python
"""Make graphics and example image for coordinate tutorial

Expects MNI nonlinear template t1 and t2 images in directory of script -
specifically these files:

* mni_icbm152_t1_tal_nlin_asym_09a.nii
* mni_icbm152_t2_tal_nlin_asym_09a.nii

Requires nipy and matplotlib.

Executing this script generates the following files in the current directory:

* localizer.png (pretend localizer sagittal image)
* someones_epi.nii.gz (pretend single EPI volume)
* someones_anatomy.nii.gz (pretend single subject structural)
"""

import math

import matplotlib.pyplot as plt
import nipy
import nipy.algorithms.resample as rsm
import nipy.core.api as nca
import numpy as np
import numpy.linalg as npl

import nibabel.eulerangles as euler

T1_IMG = 'mni_icbm152_t1_tal_nlin_asym_09a.nii'
T2_IMG = 'mni_icbm152_t2_tal_nlin_asym_09a.nii'

imgs = []
for img_fname in (T1_IMG, T2_IMG):
    img = nipy.load_image(img_fname)
    # Set affine as for FOV, not AC
    RZS = img.affine[:3, :3]
    vox_fov_center = -(np.array(img.shape) - 1) / 2.0
    T = RZS.dot(vox_fov_center)
    img.affine[:3, 3] = T
    # Take stuff off the top of the full image, to emphasize FOV
    img_z_shave = 10
    # Take stuff off left and right to save disk space
    img_x_shave = 20
    img = img[img_x_shave:-img_x_shave, :, :-img_z_shave]
    imgs.append(img)

t1_img, t2_img = imgs

# Make fake localizer
data = t1_img.get_fdata()
n_x, n_y, n_z = img.shape
mid_x = round(n_x / 2)

sagittal = data[mid_x, :, :].T

# EPI bounding box
# 3 points on a not-completely-rectangular box. The box is to give a by-eye
# estimate, then we work out the box side lengths and make a rectangular box
# from those, using the origin point
epi_bl = np.array((20, 15)) * 2
epi_br = np.array((92, 70)) * 2
epi_tl = np.array((7, 63)) * 2
# Find lengths of sides
epi_y_len = np.sqrt((np.subtract(epi_bl, epi_tl) ** 2).sum())
epi_x_len = np.sqrt((np.subtract(epi_bl, epi_br) ** 2).sum())
x, y = 0, 1
# Make a rectangular box with these sides


def make_ortho_box(bl, x_len, y_len):
    """Make a box with sides parallel to the axes"""
    return np.array(
        (bl, [bl[x] + x_len, bl[y]], [bl[x], bl[y] + y_len], [bl[x] + x_len, bl[y] + y_len])
    )


orth_epi_box = make_ortho_box(epi_bl, epi_x_len, epi_y_len)

# Structural bounding box
anat_bl = (25, 3)
anat_x_len = 185
anat_y_len = 155
anat_box = make_ortho_box(anat_bl, anat_x_len, anat_y_len)


def plot_line(pt1, pt2, fmt='r-', label=None):
    plt.plot([pt1[0], pt2[0]], [pt1[1], pt2[1]], fmt, label=label)


def plot_box(box_def, fmt='r-', label=None):
    bl, br, tl, tr = box_def
    plot_line(bl, br, fmt, label=label)
    plot_line(bl, tl, fmt)
    plot_line(br, tr, fmt)
    plot_line(tl, tr, fmt)


def rotate_box(box_def, angle, origin):
    origin = np.atleast_2d(origin)
    box_def_zeroed = box_def - origin
    cost = math.cos(angle)
    sint = math.sin(angle)
    rot_array = np.array([[cost, -sint], [sint, cost]])
    box_def_zeroed = np.dot(rot_array, box_def_zeroed.T).T
    return box_def_zeroed + origin


def labeled_point(pt, marker, text, markersize=10, color='k'):
    plt.plot(pt[0], pt[1], marker, markersize=markersize)
    plt.text(pt[0] + markersize / 2, pt[1] - markersize / 2, text, color=color)


def plot_localizer():
    plt.imshow(sagittal, cmap='gray', origin='lower', extent=sag_extents)
    plt.xlabel('mm from isocenter')
    plt.ylabel('mm from isocenter')


def save_plot():
    # Plot using global variables
    plot_localizer()

    def vx2mm(pts):
        return pts - iso_center

    plot_box(vx2mm(rot_box), label='EPI bounding box')
    plot_box(vx2mm(anat_box), 'b-', label='Structural bounding box')
    labeled_point(vx2mm(epi_center), 'ro', 'EPI FOV center')
    labeled_point(vx2mm(anat_center), 'bo', 'Structural FOV center')
    labeled_point(vx2mm(iso_center), 'g^', 'Magnet isocenter')
    plt.axis('tight')
    plt.legend(loc='lower right')
    plt.title('Scanner localizer image')
    plt.savefig('localizer.png')


angle = 0.3
rot_box = rotate_box(orth_epi_box, angle, orth_epi_box[0])
epi_center = np.mean(rot_box, axis=0)
anat_center = np.mean(anat_box, axis=0)
# y axis on the plot is first axis of image
sag_y, sag_x = sagittal.shape
iso_center = (np.array([sag_x, sag_y]) - 1) / 2.0
sag_extents = [-iso_center[0], iso_center[0], -iso_center[1], iso_center[1]]

# Back to image coordinates
br_img = np.array([0, rot_box[0, 0], rot_box[0, 1]])
epi_trans = np.eye(4)
epi_trans[:3, 3] = -br_img
rot = np.eye(4)
rot[:3, :3] = euler.euler2mat(0, 0, -angle)
# downsample to make smaller output image
downsamp = 1 / 3
epi_scale = np.diag([downsamp, downsamp, downsamp, 1])
# template voxels to epi box image voxels
vox2epi_vox = epi_scale.dot(rot.dot(epi_trans))
# epi image voxels to mm
epi_vox2mm = t2_img.affine.dot(npl.inv(vox2epi_vox))
# downsampled image shape
epi_vox_shape = np.array([data.shape[0], epi_x_len, epi_y_len]) * downsamp
# Make sure dimensions are odd by rounding up or down
# This makes the voxel center an integer index, which is convenient
epi_vox_shape = [np.floor(d) if np.floor(d) % 2 else np.ceil(d) for d in epi_vox_shape]
# resample, preserving affine
epi_cmap = nca.vox2mni(epi_vox2mm)
epi = rsm.resample(t2_img, epi_cmap, np.eye(4), epi_vox_shape)
epi_data = epi.get_fdata()
# Do the same kind of thing for the anatomical scan
anat_vox_sizes = [2.75, 2.75, 2.75]
anat_scale = npl.inv(np.diag(anat_vox_sizes + [1]))
anat_trans = np.eye(4)
anat_trans[:3, 3] = -np.array([0, anat_box[0, 0], anat_box[0, 1]])
vox2anat_vox = anat_scale.dot(anat_trans)
anat_vox2mm = t1_img.affine.dot(npl.inv(vox2anat_vox))
anat_vox_shape = np.round(np.divide([data.shape[0], anat_x_len, anat_y_len], anat_vox_sizes))
anat_cmap = nca.vox2mni(anat_vox2mm)
anat = rsm.resample(t1_img, anat_cmap, np.eye(4), anat_vox_shape)
anat_data = anat.get_fdata()

save_plot()
nipy.save_image(epi, 'someones_epi.nii.gz', dtype_from='uint8')
nipy.save_image(anat, 'someones_anatomy.nii.gz', dtype_from='uint8')

# Do progressive transforms
epi2_vox = make_ortho_box((0, 0), epi_vox_shape[1], epi_vox_shape[2])
epi_vox_sizes = np.sqrt(np.sum(epi_vox2mm[:3, :3] ** 2, axis=0))
epi2_scaled = np.diag(epi_vox_sizes[1:]).dot(epi2_vox.T).T
epi2_rotted = rotate_box(epi2_scaled, angle, (0, 0))
epi2_pulled = epi2_rotted + epi_vox2mm[1:3, 3]
plt.figure()
plot_localizer()
plot_box(epi2_vox, 'k', label='voxels')
plot_box(epi2_scaled, 'g', label='scaled')
plot_box(epi2_rotted, 'y', label='scaled, rotated')
plot_box(epi2_pulled, 'r', label='scaled, rotated, translated')
plt.legend(loc='upper left')
plt.title('Anatomy of an affine transform')
plt.savefig('illustrating_affine.png')



"""Field class defining common header fields in tractogram files"""


class Field:
    """Header fields common to multiple streamline file formats.

    In IPython, use `nibabel.streamlines.Field??` to list them.
    """

    NB_STREAMLINES = 'nb_streamlines'
    STEP_SIZE = 'step_size'
    METHOD = 'method'
    NB_SCALARS_PER_POINT = 'nb_scalars_per_point'
    NB_PROPERTIES_PER_STREAMLINE = 'nb_properties_per_streamline'
    NB_POINTS = 'nb_points'
    VOXEL_SIZES = 'voxel_sizes'
    DIMENSIONS = 'dimensions'
    MAGIC_NUMBER = 'magic_number'
    ORIGIN = 'origin'
    VOXEL_TO_RASMM = 'voxel_to_rasmm'
    VOXEL_ORDER = 'voxel_order'
    ENDIANNESS = 'endianness'



"""Testing dataobj_images module"""

import numpy as np

from nibabel.dataobj_images import DataobjImage
from nibabel.filebasedimages import FileBasedHeader
from nibabel.tests.test_filebasedimages import TestFBImageAPI as _TFI
from nibabel.tests.test_image_api import DataInterfaceMixin


class DoNumpyImage(DataobjImage):
    header_class = FileBasedHeader
    valid_exts = ('.npy',)
    files_types = (('image', '.npy'),)

    @classmethod
    def from_file_map(klass, file_map, mmap=True, keep_file_open=None):
        if mmap not in (True, False, 'c', 'r'):
            raise ValueError("mmap should be one of {True, False, 'c', 'r'}")
        if mmap is True:
            mmap = 'c'
        elif mmap is False:
            mmap = None
        with file_map['image'].get_prepare_fileobj('rb') as fobj:
            try:
                arr = np.load(fobj, mmap=mmap)
            except:
                arr = np.load(fobj)
        return klass(arr)

    def to_file_map(self, file_map=None):
        file_map = self.file_map if file_map is None else file_map
        with file_map['image'].get_prepare_fileobj('wb') as fobj:
            np.save(fobj, self.dataobj)

    def get_data_dtype(self):
        return self.dataobj.dtype

    def set_data_dtype(self, dtype):
        self._dataobj = self._dataobj.astype(dtype)


class TestDataobjAPI(_TFI, DataInterfaceMixin):
    """Validation for DataobjImage instances"""

    # A callable returning an image from ``image_maker(data, header)``
    image_maker = DoNumpyImage



"""Look for changes in numpy behavior over versions"""

from functools import cache

import numpy as np


@cache
def memmap_after_ufunc() -> bool:
    """Return True if ufuncs on memmap arrays always return memmap arrays

    This should be True for numpy < 1.12, False otherwise.
    """
    with open(__file__, 'rb') as fobj:
        mm_arr = np.memmap(fobj, mode='r', shape=(10,), dtype=np.uint8)
        return isinstance(mm_arr + 1, np.memmap)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Thin layer around xml.etree.ElementTree, to abstract nibabel xml support"""

from io import BytesIO
from xml.etree.ElementTree import Element, SubElement, tostring  # noqa: F401
from xml.parsers.expat import ParserCreate

from .filebasedimages import FileBasedHeader


class XmlSerializable:
    """Basic interface for serializing an object to XML"""

    def _to_xml_element(self) -> Element:
        """Output should be a xml.etree.ElementTree.Element"""
        raise NotImplementedError

    def to_xml(self, enc='utf-8', **kwargs) -> bytes:
        r"""Generate an XML bytestring with a given encoding.

        Parameters
        ----------
        enc : :class:`string`
            Encoding to use for the generated bytestring. Default: 'utf-8'
        \*\*kwargs : :class:`dict`
            Additional keyword arguments to :func:`xml.etree.ElementTree.tostring`.
        """
        ele = self._to_xml_element()
        return tostring(ele, enc, **kwargs)


class XmlBasedHeader(FileBasedHeader, XmlSerializable):
    """Basic wrapper around FileBasedHeader and XmlSerializable."""


class XmlParser:
    """Base class for defining how to parse xml-based image snippets.

    Image-specific parsers should define:
        StartElementHandler
        EndElementHandler
        CharacterDataHandler
    """

    HANDLER_NAMES = ['StartElementHandler', 'EndElementHandler', 'CharacterDataHandler']

    def __init__(self, encoding='utf-8', buffer_size=35000000, verbose=0):
        """
        Parameters
        ----------
        encoding : str
            string containing xml document

        buffer_size: None or int, optional
            size of read buffer. None uses default buffer_size
            from xml.parsers.expat.

        verbose : int, optional
            amount of output during parsing (0=silent, by default).
        """
        self.encoding = encoding
        self.buffer_size = buffer_size
        self.verbose = verbose
        self.fname = None  # set on calls to parse

    def _create_parser(self):
        """Internal function that allows subclasses to mess
        with the underlying parser, if desired."""

        parser = ParserCreate(encoding=self.encoding)  # from xml package
        parser.buffer_text = True
        if self.buffer_size is not None:
            parser.buffer_size = self.buffer_size
        return parser

    def parse(self, string=None, fname=None, fptr=None):
        """
        Parameters
        ----------
        string : bytes
            string (as a bytes object) containing xml document

        fname : str
            file name of an xml document.

        fptr : file pointer
            open file pointer to an xml documents
        """
        if int(string is not None) + int(fptr is not None) + int(fname is not None) != 1:
            raise ValueError('Exactly one of fptr, fname, string must be specified.')

        if string is not None:
            fptr = BytesIO(string)
        elif fname is not None:
            fptr = open(fname)

        # store the name of the xml file in case it is needed during parsing
        self.fname = getattr(fptr, 'name', None)
        parser = self._create_parser()
        for name in self.HANDLER_NAMES:
            setattr(parser, name, getattr(self, name))
        parser.ParseFile(fptr)

    def StartElementHandler(self, name, attrs):
        raise NotImplementedError

    def EndElementHandler(self, name):
        raise NotImplementedError

    def CharacterDataHandler(self, data):
        raise NotImplementedError



"""Helper functions for tests"""

from io import BytesIO

import numpy as np

from ..optpkg import optional_package

have_scipy = optional_package('scipy.io')[1]

from numpy.testing import assert_array_equal


def bytesio_filemap(klass):
    """Return bytes io filemap for this image class `klass`"""
    file_map = klass.make_file_map()
    for fileholder in file_map.values():
        fileholder.fileobj = BytesIO()
        fileholder.pos = 0
    return file_map


def bytesio_round_trip(img):
    """Save then load image from bytesio"""
    klass = img.__class__
    bytes_map = bytesio_filemap(klass)
    img.to_file_map(bytes_map)
    return klass.from_file_map(bytes_map)


def assert_data_similar(arr, params):
    """Check data is the same if recorded, otherwise check summaries

    Helper function to test image array data `arr` against record in `params`,
    where record can be the array itself, or summary values from the array.

    Parameters
    ----------
    arr : array-like
        Something that results in an array after ``np.asarry(arr)``
    params : mapping
        Mapping that has either key ``data`` with value that is array-like, or
        key ``data_summary`` with value a dict having keys ``min``, ``max``,
        ``mean``
    """
    if 'data' in params:
        assert_array_equal(arr, params['data'])
        return
    summary = params['data_summary']
    real_arr = np.asarray(arr)
    assert np.allclose(
        (real_arr.min(), real_arr.max(), real_arr.mean()),
        (summary['min'], summary['max'], summary['mean']),
    )



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from os.path import join as pjoin

import numpy as np
import pytest

from .. import minc2
from ..minc2 import Minc2File, Minc2Image
from ..optpkg import optional_package
from ..testing import data_path
from . import test_minc1 as tm2

h5py, have_h5py, setup_module = optional_package('h5py')

# Example images in format expected for ``test_image_api``, adding ``zooms``
# item.
EXAMPLE_IMAGES = [
    dict(
        fname=pjoin(data_path, 'small.mnc'),
        shape=(18, 28, 29),
        dtype=np.int16,
        affine=np.array(
            [
                [0, 0, 7.0, -98],
                [0, 8.0, 0, -134],
                [9.0, 0, 0, -72],
                [0, 0, 0, 1],
            ]
        ),
        zooms=(9.0, 8.0, 7.0),
        # These values from mincstats
        data_summary=dict(min=0.1185331417, max=92.87690699, mean=31.2127952),
        is_proxy=True,
    ),
    dict(
        fname=pjoin(data_path, 'minc2_1_scale.mnc'),
        shape=(10, 20, 20),
        dtype=np.uint8,
        affine=np.array(
            [
                [0, 0, 2.0, -20],
                [0, 2.0, 0, -20],
                [2.0, 0, 0, -10],
                [0, 0, 0, 1],
            ]
        ),
        zooms=(2.0, 2.0, 2.0),
        # These values from mincstats
        data_summary=dict(min=0.2082842439, max=0.2094327615, mean=0.2091292083),
        is_proxy=True,
    ),
    dict(
        fname=pjoin(data_path, 'minc2_4d.mnc'),
        shape=(2, 10, 20, 20),
        dtype=np.uint8,
        affine=np.array(
            [
                [0, 0, 2.0, -20],
                [0, 2.0, 0, -20],
                [2.0, 0, 0, -10],
                [0, 0, 0, 1],
            ]
        ),
        zooms=(1.0, 2.0, 2.0, 2.0),
        # These values from mincstats
        data_summary=dict(min=0.2078431373, max=1.498039216, mean=0.9090422837),
        is_proxy=True,
    ),
    dict(
        fname=pjoin(data_path, 'minc2-no-att.mnc'),
        shape=(10, 20, 20),
        dtype=np.uint8,
        affine=np.array(
            [
                [0, 0, 1.0, 0],
                [0, 1.0, 0, 0],
                [1.0, 0, 0, 0],
                [0, 0, 0, 1],
            ]
        ),
        zooms=(1.0, 1.0, 1.0),
        # These values from SPM2/mincstats
        data_summary=dict(min=0.20784314, max=0.74901961, mean=0.6061103),
        is_proxy=True,
    ),
    dict(
        fname=pjoin(data_path, 'minc2-4d-d.mnc'),
        shape=(5, 16, 16, 16),
        dtype=np.float64,
        affine=np.array(
            [
                [1.0, 0.0, 0.0, -6.96],
                [0.0, 1.0, 0.0, -12.453],
                [0.0, 0.0, 1.0, -9.48],
                [0.0, 0.0, 0.0, 1.0],
            ]
        ),
        zooms=(1.0, 1.0, 1.0, 1.0),
        # These values from mincstats
        data_summary=dict(min=0.0, max=5.0, mean=2.00078125),
        is_proxy=True,
    ),
]

if have_h5py:

    class TestMinc2File(tm2._TestMincFile):
        module = minc2
        file_class = Minc2File
        opener = h5py.File
        test_files = EXAMPLE_IMAGES

    class TestMinc2Image(tm2.TestMinc1Image):
        image_class = Minc2Image
        eg_images = (pjoin(data_path, 'small.mnc'),)
        module = minc2


def test_bad_diminfo():
    fname = pjoin(data_path, 'minc2_baddim.mnc')
    # File has a bad spacing field 'xspace' when it should be
    # `irregular`, `regular__` or absent (default to regular__).
    # We interpret an invalid spacing as absent, but warn.
    with pytest.warns(UserWarning):
        Minc2Image.from_filename(fname)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Tests for data module"""

import os
import sys
import tempfile
from os import environ as env
from os.path import join as pjoin
from tempfile import TemporaryDirectory

import pytest

from .. import data as nibd
from ..data import (
    Bomber,
    DataError,
    Datasource,
    VersionedDatasource,
    _cfg_value,
    datasource_or_bomber,
    find_data_dir,
    get_data_path,
    make_datasource,
)
from .test_environment import DATA_KEY, USER_KEY, with_environment  # noqa: F401


@pytest.fixture
def with_nimd_env(request, with_environment):  # noqa: F811
    DATA_FUNCS = {}
    DATA_FUNCS['home_dir_func'] = nibd.get_nipy_user_dir
    DATA_FUNCS['sys_dir_func'] = nibd.get_nipy_system_dir
    DATA_FUNCS['path_func'] = nibd.get_data_path
    yield
    nibd.get_nipy_user_dir = DATA_FUNCS['home_dir_func']
    nibd.get_nipy_system_dir = DATA_FUNCS['sys_dir_func']
    nibd.get_data_path = DATA_FUNCS['path_func']


def test_datasource():
    # Tests for DataSource
    pth = pjoin('some', 'path')
    ds = Datasource(pth)
    assert ds.get_filename('unlikeley') == pjoin(pth, 'unlikeley')
    assert ds.get_filename('un', 'like', 'ley') == pjoin(pth, 'un', 'like', 'ley')


def test_versioned():
    with TemporaryDirectory() as tmpdir:
        with pytest.raises(DataError):
            VersionedDatasource(tmpdir)
        tmpfile = pjoin(tmpdir, 'config.ini')
        # ini file, but wrong section
        with open(tmpfile, 'w') as fobj:
            fobj.write('[SOMESECTION]\n')
            fobj.write('version = 0.1\n')
        with pytest.raises(DataError):
            VersionedDatasource(tmpdir)
        # ini file, but right section, wrong key
        with open(tmpfile, 'w') as fobj:
            fobj.write('[DEFAULT]\n')
            fobj.write('somekey = 0.1\n')
        with pytest.raises(DataError):
            VersionedDatasource(tmpdir)
        # ini file, right section and key
        with open(tmpfile, 'w') as fobj:
            fobj.write('[DEFAULT]\n')
            fobj.write('version = 0.1\n')
        vds = VersionedDatasource(tmpdir)
        assert vds.version == '0.1'
        assert vds.version_no == 0.1
        assert vds.major_version == 0
        assert vds.minor_version == 1
        assert vds.get_filename('config.ini') == tmpfile
        # ini file, right section and key, funny value
        with open(tmpfile, 'w') as fobj:
            fobj.write('[DEFAULT]\n')
            fobj.write('version = 0.1.2.dev\n')
        vds = VersionedDatasource(tmpdir)
        assert vds.version == '0.1.2.dev'
        assert vds.version_no == 0.1
        assert vds.major_version == 0
        assert vds.minor_version == 1


def test__cfg_value():
    # no file, return ''
    assert _cfg_value('/implausible_file') == ''
    # try files
    try:
        fd, tmpfile = tempfile.mkstemp()
        fobj = os.fdopen(fd, 'wt')
        # wrong section, right key
        fobj.write('[strange section]\n')
        fobj.write('path = /some/path\n')
        fobj.flush()
        assert _cfg_value(tmpfile) == ''
        # right section, wrong key
        fobj.write('[DATA]\n')
        fobj.write('funnykey = /some/path\n')
        fobj.flush()
        assert _cfg_value(tmpfile) == ''
        # right section, right key
        fobj.write('path = /some/path\n')
        fobj.flush()
        assert _cfg_value(tmpfile) == '/some/path'
        fobj.close()
    finally:
        try:
            os.unlink(tmpfile)
        except:
            pass


def test_data_path(with_nimd_env):
    # wipe out any sources of data paths
    if DATA_KEY in env:
        del env[DATA_KEY]
    if USER_KEY in env:
        del os.environ[USER_KEY]
    fake_user_dir = '/user/path'
    nibd.get_nipy_system_dir = lambda: '/unlikely/path'
    nibd.get_nipy_user_dir = lambda: fake_user_dir
    # now we should only have anything pointed to in the user's dir
    old_pth = get_data_path()
    # We should have only sys.prefix and, iff sys.prefix == /usr,
    # '/usr/local'.  This last to is deal with Debian patching to
    # distutils.
    def_dirs = [pjoin(sys.prefix, 'share', 'nipy')]
    if sys.prefix == '/usr':
        def_dirs.append(pjoin('/usr/local', 'share', 'nipy'))
    assert old_pth == def_dirs + ['/user/path']
    # then we'll try adding some of our own
    tst_pth = '/a/path' + os.path.pathsep + '/b/ path'
    tst_list = ['/a/path', '/b/ path']
    # First, an environment variable
    os.environ[DATA_KEY] = tst_list[0]
    assert get_data_path() == tst_list[:1] + old_pth
    os.environ[DATA_KEY] = tst_pth
    assert get_data_path() == tst_list + old_pth
    del os.environ[DATA_KEY]
    # Next, make a fake user directory, and put a file in there
    with TemporaryDirectory() as tmpdir:
        tmpfile = pjoin(tmpdir, 'config.ini')
        with open(tmpfile, 'w') as fobj:
            fobj.write('[DATA]\n')
            fobj.write(f'path = {tst_pth}')
        nibd.get_nipy_user_dir = lambda: tmpdir
        assert get_data_path() == tst_list + def_dirs + [tmpdir]
    nibd.get_nipy_user_dir = lambda: fake_user_dir
    assert get_data_path() == old_pth
    # with some trepidation, the system config files
    with TemporaryDirectory() as tmpdir:
        nibd.get_nipy_system_dir = lambda: tmpdir
        tmpfile = pjoin(tmpdir, 'an_example.ini')
        with open(tmpfile, 'w') as fobj:
            fobj.write('[DATA]\n')
            fobj.write(f'path = {tst_pth}\n')
        tmpfile = pjoin(tmpdir, 'another_example.ini')
        with open(tmpfile, 'w') as fobj:
            fobj.write('[DATA]\n')
            fobj.write('path = {}\n'.format('/path/two'))
        assert get_data_path() == tst_list + ['/path/two'] + old_pth


def test_find_data_dir():
    here, fname = os.path.split(__file__)
    # here == '<rootpath>/nipy/utils/tests'
    under_here, subhere = os.path.split(here)
    # under_here == '<rootpath>/nipy/utils'
    # subhere = 'tests'
    # fails with non-existent path
    with pytest.raises(DataError):
        find_data_dir([here], 'implausible', 'directory')
    # fails with file, when directory expected
    with pytest.raises(DataError):
        find_data_dir([here], fname)
    # passes with directory that exists
    dd = find_data_dir([under_here], subhere)
    assert dd == here
    # and when one path in path list does not work
    dud_dir = pjoin(under_here, 'implausible')
    dd = find_data_dir([dud_dir, under_here], subhere)
    assert dd == here


def test_make_datasource(with_nimd_env):
    pkg_def = dict(relpath='pkg')
    with TemporaryDirectory() as tmpdir:
        nibd.get_data_path = lambda: [tmpdir]
        with pytest.raises(DataError):
            make_datasource(pkg_def)
        pkg_dir = pjoin(tmpdir, 'pkg')
        os.mkdir(pkg_dir)
        with pytest.raises(DataError):
            make_datasource(pkg_def)
        tmpfile = pjoin(pkg_dir, 'config.ini')
        with open(tmpfile, 'w') as fobj:
            fobj.write('[DEFAULT]\n')
            fobj.write('version = 0.1\n')
        ds = make_datasource(pkg_def, data_path=[tmpdir])
        assert ds.version == '0.1'


@pytest.mark.xfail(raises=DataError)
def test_bomber():
    b = Bomber('bomber example', 'a message')
    b.any_attribute  # no error


def test_bomber_inspect():
    b = Bomber('bomber example', 'a message')
    assert not hasattr(b, 'any_attribute')


def test_datasource_or_bomber(with_nimd_env):
    pkg_def = dict(relpath='pkg')
    with TemporaryDirectory() as tmpdir:
        nibd.get_data_path = lambda: [tmpdir]
        ds = datasource_or_bomber(pkg_def)
        with pytest.raises(DataError):
            ds.get_filename('some_file.txt')
        pkg_dir = pjoin(tmpdir, 'pkg')
        os.mkdir(pkg_dir)
        tmpfile = pjoin(pkg_dir, 'config.ini')
        with open(tmpfile, 'w') as fobj:
            fobj.write('[DEFAULT]\n')
            fobj.write('version = 0.2\n')
        ds = datasource_or_bomber(pkg_def)
        ds.get_filename('some_file.txt')
        # check that versioning works
        pkg_def['min version'] = '0.2'
        ds = datasource_or_bomber(pkg_def)  # OK
        ds.get_filename('some_file.txt')
        pkg_def['min version'] = '0.3'
        ds = datasource_or_bomber(pkg_def)  # not OK
        with pytest.raises(DataError):
            ds.get_filename('some_file.txt')



"""Descriptor support for NIPY

Utilities to support special Python descriptors [1,2], in particular
:func:`~functools.cached_property`, which has been available in the Python
standard library since Python 3.8. We currently maintain aliases from
earlier names for this descriptor, specifically `OneTimeProperty` and `auto_attr`.

:func:`~functools.cached_property` creates properties that are computed once
and then stored as regular attributes. They can thus be evaluated
later in the object's life cycle, but once evaluated they become normal, static
attributes with no function call overhead on access or any other constraints.

A special ResetMixin class is provided to add a .reset() method to users who
may want to have their objects capable of resetting these computed properties
to their 'untriggered' state.

References
----------
[1] How-To Guide for Descriptors, Raymond
Hettinger. https://docs.python.org/howto/descriptor.html

[2] Python data model, https://docs.python.org/reference/datamodel.html
"""

from __future__ import annotations

from functools import cached_property

from nibabel.deprecated import deprecate_with_version

# -----------------------------------------------------------------------------
# Classes and Functions
# -----------------------------------------------------------------------------


class ResetMixin:
    """A Mixin class to add a .reset() method to users of cached_property.

    By default, cached properties, once computed, become static.  If they happen
    to depend on other parts of an object and those parts change, their values
    may now be invalid.

    This class offers a .reset() method that users can call *explicitly* when
    they know the state of their objects may have changed and they want to
    ensure that *all* their special attributes should be invalidated.  Once
    reset() is called, all their cached properties are reset to their
    :func:`~functools.cached_property` descriptors,
    and their accessor functions will be triggered again.

    .. warning::

       If a class has a set of attributes that are cached_property, but that
       can be initialized from any one of them, do NOT use this mixin!  For
       instance, UniformTimeSeries can be initialized with only sampling_rate
       and t0, sampling_interval and time are auto-computed.  But if you were
       to reset() a UniformTimeSeries, it would lose all 4, and there would be
       then no way to break the circular dependency chains.

       If this becomes a problem in practice (for our analyzer objects it
       isn't, as they don't have the above pattern), we can extend reset() to
       check for a _no_reset set of names in the instance which are meant to be
       kept protected.  But for now this is NOT done, so caveat emptor.

    Examples
    --------

    >>> class A(ResetMixin):
    ...     def __init__(self,x=1.0):
    ...         self.x = x
    ...
    ...     @cached_property
    ...     def y(self):
    ...         print('*** y computation executed ***')
    ...         return self.x / 2.0

    >>> a = A(10)

    About to access y twice, the second time no computation is done:

    >>> a.y
    *** y computation executed ***
    5.0
    >>> a.y
    5.0

    Changing x

    >>> a.x = 20

    a.y doesn't change to 10, since it is a static attribute:

    >>> a.y
    5.0

    We now reset a, and this will then force all auto attributes to recompute
    the next time we access them:

    >>> a.reset()

    About to access y twice again after reset():

    >>> a.y
    *** y computation executed ***
    10.0
    >>> a.y
    10.0
    """

    def reset(self) -> None:
        """Reset all cached_property attributes that may have fired already."""
        # To reset them, we simply remove them from the instance dict.  At that
        # point, it's as if they had never been computed.  On the next access,
        # the accessor function from the parent class will be called, simply
        # because that's how the python descriptor protocol works.
        for mname, mval in self.__class__.__dict__.items():
            if mname in self.__dict__ and isinstance(mval, cached_property):
                delattr(self, mname)


OneTimeProperty = cached_property
auto_attr = cached_property

# -----------------------------------------------------------------------------
# Deprecated API
# -----------------------------------------------------------------------------

# For backwards compatibility
setattr_on_read = deprecate_with_version(
    message='setattr_on_read has been renamed to auto_attr. Please use nibabel.onetime.auto_attr',
    since='3.2',
    until='5.0',
)(auto_attr)
